#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""AnaSimpleViewer2"""

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL license version 2 under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL license version 2 as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license version 2 and that you accept its terms.

from __future__ import absolute_import, print_function

import os
import sys
import time

import numpy as np
import PyQt5
import six
from PyQt5.QtGui import QColor, QIcon, QLabel, QWidget
from PyQt5.QtWidgets import QMessageBox
from six.moves import zip
from soma import aims
from soma.aims import colormaphints
from soma.qt_gui import qt_backend
from soma.qt_gui.qt_backend import Qt, QtCore
from soma.qt_gui.qt_backend.uic import loadUi

from populse_mia.software_properties import Config
from populse_mia.user_interface.data_viewer.anatomist_2.snd_window import (
    NewWindowViewer,
)

try:
    import anatomist.direct.api as ana
except ImportError:
    print(
        "\nAnatomist seems not to be installed. The data_viewer anatomist "
        "and anatomist_2 will not work...\n"
    )
# the following imports have to be made after the qApp.startingUp() test
# since they do instantiate Anatomist for registry to work.
try:
    from anatomist.cpp.simplecontrols import (
        Simple2DControl,
        registerSimpleControls,
    )
except ImportError:
    print(
        "\nAnatomist seems not to be installed. The data_viewer anatomist "
        "and anatomist_2 will not work...\n"
    )
# determine whether we are using Qt4 or Qt5, and hack a little bit accordingly
# the boolean qt4 global variable will tell it for later usage
qt_backend.set_qt_backend(compatible_qt5=True)


class LeftSimple3DControl(Simple2DControl):
    """
    define another control where rotation is with the left mouse button
    (useful for touch devices)
    """

    def __init__(self, prio=25, name="LeftSimple3DControl"):
        """blabla"""
        super(LeftSimple3DControl, self).__init__(prio, name)

    def eventAutoSubscription(self, pool):
        """blabla"""
        key = QtCore.Qt
        NoModifier = key.NoModifier
        ControlModifier = key.ControlModifier
        super(LeftSimple3DControl, self).eventAutoSubscription(pool)
        self.mouseLongEventUnsubscribe(key.LeftButton, NoModifier)
        self.mouseLongEventSubscribe(
            key.LeftButton,
            NoModifier,
            pool.action("ContinuousTrackball").beginTrackball,
            pool.action("ContinuousTrackball").moveTrackball,
            pool.action("ContinuousTrackball").endTrackball,
            True,
        )
        self.keyPressEventSubscribe(
            key.Key_Space,
            ControlModifier,
            pool.action("ContinuousTrackball").startOrStop,
        )
        self.mousePressButtonEventSubscribe(
            key.MiddleButton, NoModifier, pool.action("LinkAction").execLink
        )


class VolRenderControl(LeftSimple3DControl):
    """
    define another control where cut slice rotation is with the middle mouse
    button
    """

    def __init__(self, prio=25, name="VolRenderControl"):
        """blabla"""
        super(VolRenderControl, self).__init__(prio, name)

    def eventAutoSubscription(self, pool):
        """blabla"""
        super(VolRenderControl, self).eventAutoSubscription(pool)
        self.mouseLongEventUnsubscribe(Qt.Qt.MiddleButton, Qt.Qt.NoModifier)
        self.mouseLongEventSubscribe(
            Qt.Qt.MiddleButton,
            Qt.Qt.NoModifier,
            pool.action("TrackCutAction").beginTrackball,
            pool.action("TrackCutAction").moveTrackball,
            pool.action("TrackCutAction").endTrackball,
            True,
        )


class AnaSimpleViewer2(Qt.QObject):
    """
    AnaSimpleViewer is a "simple viewer" application and widget, which can be
    used using the "anasimpleviewer.py" command, or included in a custom widget
    as a library module.

    It includes an objects list and 4 3D views (anatomist windows). Objects
    loaded are added in all views, and can be hidden or shown using the "add"
    and "remove" buttons.

    The AnaSimpleViewer class holds methods for menu/actions callbacks, and
    utility functions like load/view objects, remove/delete, etc.

    It is a QObject, but not a QWidget. The widget can be accessed as the
    ``awidget`` attribute in the AnaSimpleViewer instance.

    As it is more intended to be used as a complete application, and it is
    simpler to handle in Anatomist, some global Anatomist config variables and
    controls may be set within AnaSimpleViewer. This is done optionally using
    the :meth:`init_global_handlers` method, which is called by the constructor
    if the argument `init_global_handlers` is not set to False when calling it.
    """

    _global_handlers_initialized = False

    def __init__(self, init_global_handlers=None):
        """blabla"""
        Qt.QObject.__init__(self)

        a = ana.Anatomist("-b")

        if init_global_handlers:
            self.init_global_handlers()

        # ui file for dataviewer anasimpleviewer_2
        uifile = "mainwindow.ui"
        cwd = os.getcwd()
        mainwindowdir = os.path.dirname(__file__)
        os.chdir(mainwindowdir)
        awin = loadUi(os.path.join(mainwindowdir, uifile))
        os.chdir(cwd)
        self.awidget = awin

        # new window popup for objects
        self.newWindow = NewWindowViewer()

        # connect GUI actions callbacks
        def findChild(x, y):
            """blabla"""
            return Qt.QObject.findChild(x, QtCore.QObject, y)

        # findChild(awin,
        #           'actionprint_view').triggered.connect(self.addNewView)
        findChild(awin, "actionTimeRunning").triggered.connect(
            self.automaticRunning
        )
        findChild(awin, "fileOpenAction").triggered.connect(self.fileOpen)
        findChild(awin, "fileExitAction").triggered.connect(self.closeAll)
        findChild(awin, "editAddAction").triggered.connect(self.editAdd)
        findChild(awin, "editRemoveAction").triggered.connect(self.editRemove)
        findChild(awin, "editDeleteAction").triggered.connect(self.editDelete)
        findChild(awin, "viewEnable_Volume_RenderingAction").toggled.connect(
            self.enableVolumeRendering
        )
        findChild(awin, "viewOpen_Anatomist_main_window").triggered.connect(
            self.open_anatomist_main_window
        )
        # manually entered coords
        le = findChild(awin, "coordXEdit")
        le.setValidator(Qt.QDoubleValidator(le))
        le = findChild(awin, "coordYEdit")
        le.setValidator(Qt.QDoubleValidator(le))
        le = findChild(awin, "coordZEdit")
        le.setValidator(Qt.QDoubleValidator(le))
        le = findChild(awin, "coordTEdit")
        le.setValidator(Qt.QDoubleValidator(le))
        del le
        findChild(awin, "coordXEdit").editingFinished.connect(
            self.coordsChanged
        )
        findChild(awin, "coordYEdit").editingFinished.connect(
            self.coordsChanged
        )
        findChild(awin, "coordZEdit").editingFinished.connect(
            self.coordsChanged
        )
        findChild(awin, "coordTEdit").editingFinished.connect(
            self.coordsChanged
        )

        objects_list = findChild(self.awidget, "objectslist")
        objects_list.setContextMenuPolicy(Qt.Qt.CustomContextMenu)
        objects_list.customContextMenuRequested.connect(self.popup_objects)

        awin.dropEvent = lambda awin, event: self.dropEvent(awin, event)
        awin.dragEnterEvent = lambda awin, event: self.dragEnterEvent(
            awin, event
        )
        awin.setAcceptDrops(True)

        self._vrenabled = False
        self.meshes2d = {}
        # register the function on the cursor notifier of anatomist. It will be
        # called when the user clicks on a window
        a.onCursorNotifier.add(self.clickHandler)

        # viewWindow: parent widget for anatomist windows
        self.viewWindow = findChild(awin, "windows")

        self.viewgridlay = Qt.QHBoxLayout(self.viewWindow)
        self.combobox = Qt.QComboBox()
        self.slider = Qt.QSlider(Qt.Qt.Horizontal)
        self.fdialog = None
        self.awindows = []
        self.aobjects = []
        self.fusion2d = []
        self.volrender = None
        self.control_3d_type = "LeftSimple3DControl"
        self.viewButtons = [
            findChild(awin, "actionAxial"),
            findChild(awin, "actionSagittal"),
            findChild(awin, "actionCoronal"),
            findChild(awin, "action3D"),
        ]
        self.displayedObjects = []
        self.files = []
        self.available_palettes = [
            "B-W_LINEAR",
            "Yellow-red",
            "RAINBOW",
            "Yellow-Red-White-Blue-Green",
            "blue-red-bright-dark",
        ]

        for action in self.viewButtons:
            action.triggered.connect(self.newDisplay)
        findChild(awin, "objectslist").itemSelectionChanged.connect(
            self.disableButtons
        )

        self.setComboBox()
        self.setSlider()
        self.combobox.currentIndexChanged.connect(self.newPalette)
        self.slider.valueChanged.connect(self.changeOpacity)

    def init_global_handlers(self):
        """
        Set some global controls / settings in Anatomist application object
        """
        if not AnaSimpleViewer2._global_handlers_initialized:
            registerSimpleControls()
            a = ana.Anatomist("-b")
            iconpath = os.path.join(str(a.anatomistSharedPath()), "icons")
            pix = Qt.QPixmap(os.path.join(iconpath, "simple3Dcontrol.png"))
            ana.cpp.IconDictionary.instance().addIcon(
                "LeftSimple3DControl", pix
            )
            ana.cpp.IconDictionary.instance().addIcon("VolRenderControl", pix)
            del pix, iconpath
            cd = ana.cpp.ControlDictionary.instance()
            cd.addControl("LeftSimple3DControl", LeftSimple3DControl, 25)
            cd.addControl("VolRenderControl", VolRenderControl, 25)

            # tweak: override some user config options
            a.config()["windowSizeFactor"] = 1.0
            a.config()["commonScannerBasedReferential"] = 1

            # register controls
            cm = ana.cpp.ControlManager.instance()
            cm.addControl("QAGLWidget3D", "", "Simple2DControl")
            cm.addControl("QAGLWidget3D", "", "LeftSimple3DControl")
            cm.addControl("QAGLWidget3D", "", "VolRenderControl")
            print("controls registered.")

            del cm

            a.setGraphParams(label_attribute="label")

            AnaSimpleViewer2._global_handlers_initialized = True

    def setComboBox(self):
        """
        Inserts a drop down menu in the toolbar. This menu contains defined
        available color palettes. The default color palettes are in
        self.available_palettes.
        """

        toolBar = Qt.QObject.findChild(self.awidget, QtCore.QObject, "toolBar")
        actionAutoRunning = Qt.QObject.findChild(
            self.awidget, QtCore.QObject, "actionTimeRunning"
        )
        label = QLabel("Palette: ")
        label.setToolTip("Change color palette of selected object")

        for palette in self.available_palettes:
            self.combobox.addItem(palette)

        toolBar.insertWidget(actionAutoRunning, label)
        toolBar.insertWidget(actionAutoRunning, self.combobox)
        sources_images_dir = Config().getSourceImageDir()

        for i in range(len(self.available_palettes)):
            icon = QIcon(
                os.path.join(sources_images_dir, self.available_palettes[i])
            )
            self.combobox.setItemIcon(i, icon)

        size = PyQt5.QtCore.QSize(200, 15)
        self.combobox.setIconSize(size)

    def setSlider(self):
        """
        Inserts opacity slider in the toolbar
        Minimum returned value is 0 and maximum value is 100
        """

        toolBar = Qt.QObject.findChild(self.awidget, QtCore.QObject, "toolBar")
        actionAutoRunning = Qt.QObject.findChild(
            self.awidget, QtCore.QObject, "actionTimeRunning"
        )
        space = QWidget().resize(5, 0)
        label = QLabel("Opacity: ")
        label.setToolTip("Change opacity of selected object")
        size = PyQt5.QtCore.QSize(120, 15)
        self.slider.setMaximumSize(size)
        self.slider.setValue(100)
        toolBar.insertWidget(actionAutoRunning, space)
        toolBar.insertWidget(actionAutoRunning, label)
        toolBar.insertWidget(actionAutoRunning, self.slider)

    def changeOpacity(self):
        """
        Changes opacity of selected object (to be more precise it changes the
        mixing rate between objects when multiple ones are displayed)
        """

        if self.selectedObjects() and len(self.displayedObjects) == 1:
            diffuse_vect = self.selectedObjects()[0].getInfo()["material"][
                "diffuse"
            ]
            self.selectedObjects()[0].setMaterial(
                diffuse=[
                    diffuse_vect[0],
                    diffuse_vect[1],
                    diffuse_vect[2],
                    self.slider.value() / 100,
                ]
            )

        elif self.selectedObjects():
            index = self.displayedObjects.index(self.selectedObjects()[0])
            a = ana.Anatomist("-b")

            if index == 0:
                a.execute(
                    "TexturingParams",
                    objects=self.fusion2d,
                    texture_index=1,
                    rate=self.slider.value() / 100,
                    mode="linear_B_if_A_black",
                )

            else:
                corrected_val = abs(self.slider.value() - 100)
                a.execute(
                    "TexturingParams",
                    objects=self.fusion2d,
                    texture_index=index,
                    rate=corrected_val / 100,
                    mode="linear_A_if_B_black",
                )

    def newPalette(self):
        """
        Sets chosen color palette in the toolbar drop down menu to
        selected object.
        """
        color = self.combobox.currentText()
        if self.selectedObjects():
            self.selectedObjects()[0].setPalette(palette=color)

    def setColorPalette(self):
        """
        Checks color palette of a selected object, displays it in the toolbar
        drop-down menu and adds it if it isn't already stored in
        self.available_palettes. If corresponding palette image exists it is
        added, otherwise only the name of the palette appears.
        """

        color = self.combobox.currentText()

        if self.selectedObjects():
            actual_pal = self.selectedObjects()[0].getInfo()["palette"][
                "palette"
            ]

            if actual_pal != color:
                if actual_pal in self.available_palettes:
                    self.combobox.setCurrentText(actual_pal)

                else:
                    self.combobox.addItem(actual_pal)
                    self.combobox.setCurrentText(actual_pal)
                    self.available_palettes.append(actual_pal)
                    sources_images_dir = Config().getSourceImageDir()
                    index = self.combobox.currentIndex()
                    icon = QIcon(
                        os.path.join(
                            sources_images_dir, self.available_palettes[index]
                        )
                    )
                    self.combobox.setItemIcon(index, icon)

    def changeConfig(self, config):
        """
        change config depending on user settings
        config : string "neuro" or "radio"
        """
        a = ana.Anatomist("-b")
        a.config()["axialConvention"] = config
        self.newDisplay()

    def changeRef(self):
        """
        change referential
        ref : Boolean
        0 : World coordinates
        1 : Image referential
        """
        self.deleteObjects(self.aobjects)
        self.loadObject(self.files, config_changed=True)

    def findChild(x, y):
        """Blabla"""

        return Qt.QObject.findChild(x, QtCore.QObject, y)

    def clickHandler(self, eventName, params):
        """
        Callback for linked cursor. In volume rendering mode, it will sync
        the VR slice to the linked cursor.
        It also updates the volumes values view
        """
        a = ana.Anatomist("-b")
        pos = params["position"]
        win = params["window"]
        wref = win.getReferential()
        # display coords in MNI referential (preferably)
        tr = a.getTransformation(wref, a.mniTemplateRef)
        if tr:
            pos2 = tr.transform(pos[:3])
        else:
            pos2 = pos

        def findChild(x, y):
            """Blabla"""
            return Qt.QObject.findChild(x, QtCore.QObject, y)

        x = findChild(self.awidget, "coordXEdit")
        x.setText("%8.3f" % pos2[0])
        y = findChild(self.awidget, "coordYEdit")
        y.setText("%8.3f" % pos2[1])
        z = findChild(self.awidget, "coordZEdit")
        z.setText("%8.3f" % pos2[2])
        t = findChild(self.awidget, "coordTEdit")
        if len(pos) < 4:
            pos = pos[:3] + [0]
        t.setText("%8.3f" % pos[3])
        # display volumes values at the given position

        valbox = findChild(self.awidget, "volumesBox")
        valbox.clear()
        # (we don't use the same widget type in Qt3 and Qt4)

        valbox.setColumnCount(2)
        valbox.setHorizontalHeaderLabels(["Volume:", "Value:"])
        if len(self.fusion2d) > 1:
            valbox.setRowCount(len(self.fusion2d) - 1)
            valbox.setVerticalHeaderLabels([""] * (len(self.fusion2d) - 1))
        i = 0
        for obj in self.fusion2d[1:]:
            # retrieve volume val'radio'ue in its own coords system
            aimsv = ana.cpp.AObjectConverter.aims(obj)
            oref = obj.getReferential()
            tr = a.getTransformation(wref, oref)
            if tr:
                pos2 = tr.transform(pos[:3])
            else:
                pos2 = pos[:3]
            vs = obj.voxelSize()
            pos2 = [int(round(x / y)) for x, y in zip(pos2, vs)]
            # pos2 in in voxels, in obj coords system
            newItem = Qt.QTableWidgetItem(obj.name)
            valbox.setItem(i, 0, newItem)
            # check bounds
            if (
                pos2[0] >= 0
                and pos2[1] >= 0
                and pos2[2] >= 0
                and pos[3] >= 0
                and pos2[0] < aimsv.getSizeX()
                and pos2[1] < aimsv.getSizeY()
                and pos2[2] < aimsv.getSizeZ()
                and pos[3] < aimsv.getSizeT()
            ):
                txt = str(aimsv.value(*pos2))
            else:
                txt = ""
            newitem = Qt.QTableWidgetItem(txt)
            valbox.setItem(i, 1, newitem)
            i += 1
        valbox.resizeColumnsToContents()

        # update volume rendering when it is enabled
        if self._vrenabled and len(self.volrender) >= 1:
            clip = self.volrender[0]
            t = a.getTransformation(
                win.getReferential(), clip.getReferential()
            )
            if t is not None:
                pos = t.transform(pos[:3])
            clip.setOffset(pos[:3])
            clip.notifyObservers()

    def automaticRunning(self):
        """
        Enable automatic running of functional images
        frame rate can be changed in preferences by the user
        """

        a = ana.Anatomist("-b")
        objects = []
        im_sec = float(Config().getViewerFramerate())
        frame_rate = 1 / im_sec

        def findChild(x, y):
            """Blabla"""

            return Qt.QObject.findChild(x, QtCore.QObject, y)

        t = findChild(self.awidget, "coordTEdit")
        sources_images_dir = Config().getSourceImageDir()
        pauseIcon = QIcon(os.path.join(sources_images_dir, "pause.png"))
        playIcon = QIcon(os.path.join(sources_images_dir, "play.png"))

        for i in range(len(self.displayedObjects)):
            objects.append(
                ana.cpp.AObjectConverter.aims(
                    self.displayedObjects[i]
                ).getSizeT()
            )
        if objects:
            nb_images = np.max(objects)
        else:
            return
        list_im = list(range(0, nb_images))
        pos = [
            float(findChild(self.awidget, "coordXEdit").text()),
            float(findChild(self.awidget, "coordYEdit").text()),
            float(findChild(self.awidget, "coordZEdit").text()),
        ]
        i = int(float(findChild(self.awidget, "coordTEdit").text()))
        if i == nb_images - 1:
            i = 0
        playAction = findChild(self.awidget, "actionTimeRunning")
        while playAction.isChecked() and i < len(list_im):
            start_time = time.time()
            playAction.setIcon(pauseIcon)
            t.setText("%8.3f" % list_im[i])
            a.execute(
                "LinkedCursor",
                window=self.awindows[0],
                position=pos[:3] + [list_im[i]],
            )
            PyQt5.QtWidgets.QApplication.processEvents()
            running_time = time.time() - start_time
            if running_time > frame_rate:
                # If iteration takes to much time we don't want to
                # make it sleep any longer (happens when fusion of images)
                pass
            else:
                time.sleep(frame_rate - running_time)
            i += 1
        playAction.setIcon(playIcon)

    def createWindow(self, wintype="Axial"):
        """
        Opens a new window in the windows grid layout.
        The new window will be set in MNI referential (except 3D for now
        because of a buf in volume rendering in direct referentials), will be
        assigned the custom control, and have no menu/toolbars.
        """
        a = ana.Anatomist("-b")
        w = a.createWindow(wintype, no_decoration=True, options={"hidden": 1})
        w.setAcceptDrops(False)
        # insert in grid layout

        x = 0
        y = 0
        i = 0
        if not hasattr(self, "_winlayouts"):
            self._winlayouts = [[0, 0], [0, 0]]
        else:
            freeslot = False
            for y in (0, 1):
                for x in (0, 1):
                    i = i + 1
                    if not self._winlayouts[x][y]:
                        freeslot = True
                        break
                if freeslot:
                    break

        # in Qt4, the widget must not have a parent before calling
        # layout.addWidget
        # self.viewgridlay.addWidget(w.getInternalRep(), x, y)
        self.viewgridlay.addWidget(w.getInternalRep())
        # self.viewgridlay.addWidget(w.getInternalRep(), 0, i)
        self._winlayouts[x][y] = 1
        # keep it in anasimpleviewer list of windows
        self.awindows.append(w)
        # set custom control
        if wintype == "3D":
            a.execute("SetControl", windows=[w], control=self.control_3d_type)
        else:
            a.execute("SetControl", windows=[w], control="Simple2DControl")
            a.assignReferential(a.mniTemplateRef, w)
            # force redrawing in MNI orientation
            # (there should be a better way to do so...)
            if wintype == "Axial":
                w.muteAxial()
                print("MUTEAXIAL", w.muteAxial)
            elif wintype == "Coronal":
                w.muteCoronal()
            elif wintype == "Sagittal":
                w.muteSagittal()
            elif wintype == "Oblique":
                w.muteOblique()
        # set a black background
        a.execute(
            "WindowConfig",
            windows=[w],
            light={"background": [0.0, 0.0, 0.0, 1.0]},
            view_size=(500, 600),
        )

    def createTotalWindow(self, views):
        """
        Create the windows which will contain the views.

        views : array containing strings "axial", "sagittal", "coronal"
                and/or "3D"
        """

        for i in views:
            self.createWindow(str(i))

            if i == "3D":
                # set a cool angle of view for 3D
                a = ana.Anatomist("-b")
                a.execute(
                    "Camera",
                    windows=[self.awindows[-1]],
                    view_quaternion=[0.404603, 0.143829, 0.316813, 0.845718],
                )

        # Sets view buttons checked for first display
        if views == ["Axial", "Sagittal", "Coronal"]:
            counter = 0

            for i in views:
                self.viewButtons[counter].setChecked(True)
                counter += 1

    def deleteTotalWindow(self):
        """
        Clear windows and fusions in order to enable new display
        """
        self.awindows.clear()
        self.fusion2d.clear()
        for i in reversed(range(self.viewgridlay.count())):
            self.viewgridlay.itemAt(i).widget().deleteLater()

    def getViewsToDisplay(self):
        """
        Check which views must be displayed
        return : array with strings
        """
        views = []
        if self.viewButtons[0].isChecked():
            views.append("Axial")
        if self.viewButtons[1].isChecked():
            views.append("Sagittal")
        if self.viewButtons[2].isChecked():
            views.append("Coronal")
        if self.viewButtons[3].isChecked():
            views.append("3D")
        return views

    def viewReferential(self, object):
        """
        Set referential at the object center to visualize it
        """
        a = ana.Anatomist("-b")
        bb = object.boundingbox()
        position = (aims.Point3df(bb[1][:3]) - bb[0][:3]) / 2.0
        wrefs = [w.getReferential() for w in self.awindows]
        srefs = set([r.uuid() for r in wrefs])
        if len(srefs) != 1:
            # not all windows in the same ref
            if aims.StandardReferentials.mniTemplateReferentialID() in srefs:
                wref_id = aims.StandardReferentials.mniTemplateReferentialID()
                wref = [r for r in wrefs if r.uuid() == wref_id][0]
            elif aims.StandardReferentials.acPcReferentialID() in srefs:
                wref = a.centralReferential()
            elif (
                aims.StandardReferentials.commonScannerBasedReferentialID()
                in srefs
            ):
                wref_id = (
                    aims.StandardReferentials.commonScannerBasedReferentialID()
                )
                wref = [r for r in wrefs if r.uuid() == wref_id][0]
            else:
                wref = wrefs[0]
            for w in self.awindows:
                w.setReferential(wref)
        else:
            wref = wrefs[0]

        t = a.getTransformation(object.getReferential(), wref)
        if not t and object.getReferential() != wref:
            # try to find a scanner-based ref and connect it to MNI
            sbref = [
                r
                for r in a.getReferentials()
                if r.uuid()
                == aims.StandardReferentials.commonScannerBasedReferentialID()
            ]
            if sbref:
                sbref = sbref[0]
                t2 = a.getTransformation(object.getReferential(), sbref)
                if t2:
                    a.execute(
                        "LoadTransformation",
                        origin=sbref,
                        destination=a.mniTemplateRef,
                        matrix=[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1],
                    )
                else:
                    # otherwise we will assume the object is in the central
                    # referential.
                    a.execute(
                        "LoadTransformation",
                        origin=object.getReferential(),
                        destination=a.centralRef,
                        matrix=[0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1],
                    )
                t = a.getTransformation(
                    object.getReferential(), self.awindows[0].getReferential()
                )
        if t:
            position = t.transform(position)
        a.execute("LinkedCursor", window=self.awindows[0], position=position)
        for w in self.awindows:
            w.focusView()

    def checkviews(self):
        """
        Prevent from closing the last view opened
        Checks how many views button are enabled and if only one is,
        it disables the button
        """
        nb_views_checked = 0
        for i in range(len(self.viewButtons)):
            if self.viewButtons[i].isChecked():
                nb_views_checked += 1
        if nb_views_checked == 1:
            for i in range(len(self.viewButtons)):
                if self.viewButtons[i].isChecked():
                    self.viewButtons[i].setEnabled(False)
        else:
            for i in range(len(self.viewButtons)):
                self.viewButtons[i].setEnabled(True)

    def newDisplay(self):
        """
        New display of windows, objects and views
        """
        self.checkviews()
        self.deleteTotalWindow()
        views = self.getViewsToDisplay()
        self.createTotalWindow(views)
        for i in range(len(self.displayedObjects)):
            self.addObject(self.displayedObjects[i])
            self.viewReferential(self.displayedObjects[i])

    def loadObject(self, files, config_changed=None):
        """
        Load objects in files and display
        Only the first object of files is displayed, the others are loaded and
        added to objectlist but not displayed
        """
        a = ana.Anatomist("-b")
        a.config()["setAutomaticReferential"] = Config().get_referential()
        a.config()["axialConvention"] = Config().getViewerConfig()

        # Progress indication
        window = Qt.QWidget()
        window.setWindowTitle("Loading data")
        window.move(800, 500)
        window.resize(300, 50)
        window.show()

        i = 0
        for fname in files:
            if fname not in self.files:
                self.files.append(fname)
            objectlist = Qt.QObject.findChild(
                self.awidget, QtCore.QObject, "objectslist"
            )
            # test if object has already been imported
            for w in range(objectlist.count()):
                if os.path.basename(fname) == objectlist.item(w).text():
                    if config_changed is True:
                        return
                    msgBox = QMessageBox()
                    msgBox.setIcon(QMessageBox.Warning)
                    msgBox.setText(
                        "Some of your objects have already been imported"
                    )
                    msgBox.setWindowTitle("Warning")
                    msgBox.setStandardButtons(QMessageBox.Ok)

                    returnValue = msgBox.exec()
                    if returnValue == QMessageBox.Ok:
                        return

            obj = a.loadObject(fname)
            if obj:
                if i == 0:
                    self.registerObject(obj)
                    i += 1
                else:
                    objectlist.addItem(obj.name)
                    # keep it in the global list
                    self.aobjects.append(obj)
                    self.colorBackgroundList()
                    if obj.objectType == "VOLUME":
                        # volume are checked for possible adequate colormaps
                        hints = colormaphints.checkVolume(
                            ana.cpp.AObjectConverter.aims(obj)
                        )
                        obj.attributed()["colormaphints"] = hints
                    i += 1

    @QtCore.Slot("anatomist::AObject *", "const std::string &")
    def objectLoaded(self, obj, filename):
        """Blabla"""

        a = ana.Anatomist("-b")
        if not obj:
            return
        o = a.AObject(a, obj)
        o.releaseAppRef()
        p = a.theProcessor()
        resetProcExec = False
        if not p.execWhileIdle():
            # allow recursive commands execution, otherwise the execute()
            # may not be done right now
            p.allowExecWhileIdle(True)
            resetProcExec = True
        self.registerObject(o)
        if resetProcExec:
            # set back recursive execution to its previous state
            p.allowExecWhileIdle(False)

    def registerObject(self, obj, views=None):
        """
        Register an object in anasimpleviewer objects list, and display it
        """
        ojectlist = Qt.QObject.findChild(
            self.awidget, QtCore.QObject, "objectslist"
        )
        ojectlist.addItem(obj.name)
        # keep it in the global list
        self.aobjects.append(obj)
        self.displayedObjects.append(obj)
        self.colorBackgroundList()

        if obj.objectType == "VOLUME":
            # volume are checked for possible adequate colormaps
            # prints the header of volume ana.cpp.AObjectConverter.aims(obj)
            hints = colormaphints.checkVolume(
                ana.cpp.AObjectConverter.aims(obj)
            )
            obj.attributed()["colormaphints"] = hints

        bb = obj.boundingbox()

        if not bb:
            # not a viewable object
            return

        # create the 4 windows if they don't exist
        if len(self.awindows) == 0:
            if views is None:
                self.createTotalWindow(["Axial", "Sagittal", "Coronal"])

            else:
                self.createTotalWindow(views)

        # view obj in these views
        self.addObject(obj)
        # set the cursor at the center of the object (actually, overcome a bug
        # in anatomist...)
        self.viewReferential(obj)

    def _displayVolume(self, obj, opts={}):
        """
        Display a volume or a Fusion2D in all windows.
        If volume rendering is allowed, 3D views will display a clipped volume
        rendering of the object.
        """
        a = ana.Anatomist("-b")
        if self._vrenabled:
            wins = [x for x in self.awindows if x.subtype() != 0]
            if len(wins) != 0:
                a.addObjects(obj, wins, **opts)
            wins = [x for x in self.awindows if x.subtype() == 0]
            if len(wins) == 0:
                return
            vr = a.fusionObjects([obj], method="VolumeRenderingFusionMethod")
            vr.releaseAppRef()
            clip = a.fusionObjects([vr], method="FusionClipMethod")
            clip.releaseAppRef()
            self.volrender = [clip, vr]
            a.addObjects(clip, wins, **opts)
        else:
            a.addObjects(obj, self.awindows, **opts)

    def addVolume(self, obj, opts={}):
        """
        Display a volume in all windows.
        If several volumes are displayed, a Fusion2D will be built to wrap all
        of them.
        If volume rendering is allowed, 3D views will display a clipped volume
        rendering of either the single volume (if only one is present), or of
        the 2D fusion.
        """
        a = ana.Anatomist("-b")

        if obj in self.fusion2d:
            return

        # hasvr = False
        if self.volrender:
            # delete the previous volume rendering
            a.deleteObjects(self.volrender)
            # hasvr = True
            self.volrender = None

        if len(self.fusion2d) == 0:
            # only one object
            self.fusion2d = [None, obj]

        else:
            # several objects: fusion them
            fusobjs = self.fusion2d[1:] + [obj]
            f2d = a.fusionObjects(fusobjs, method="Fusion2DMethod")
            f2d.releaseAppRef()

            if self.fusion2d[0] is not None:
                # destroy the previous fusion
                a.deleteObjects(self.fusion2d[0])

            else:
                a.removeObjects(self.fusion2d[1], self.awindows)

            self.fusion2d = [f2d] + fusobjs
            # repalette( fusobjs )
            obj = f2d

        if obj.objectType == "VOLUME":
            # choose a good colormap for a single volume
            if "volume_contents_likelihoods" in obj.attributed():
                cmap = colormaphints.chooseColormaps(
                    (obj.attributed()["colormaphints"],)
                )
                obj.setPalette(cmap[0])

        else:
            # choose good colormaps for the current set of volumes
            hints = [x.attributed()["colormaphints"] for x in obj.children]
            children = [
                x
                for x, y in zip(obj.children, hints)
                if "volume_contents_likelihoods" in y
            ]
            hints = [x for x in hints if "volume_contents_likelihoods" in x]
            cmaps = colormaphints.chooseColormaps(hints)

            for x, y in zip(children, cmaps):
                x.setPalette(y)

        # call a lower-level function for display and volume rendering
        self._displayVolume(obj, opts)

    def removeVolume(self, obj, opts={}):
        """
        Hides a volume from views (low-level function: use removeObject)
        """
        a = ana.Anatomist("-b")

        if obj in self.fusion2d:
            # hasvr = False

            if self.volrender:
                a.deleteObjects(self.volrender)
                self.volrender = None
                # hasvr = True

            fusobjs = [o for o in self.fusion2d[1:] if o != obj]

            if len(fusobjs) >= 2:
                f2d = a.fusionObjects(fusobjs, method="Fusion2DMethod")
                f2d.releaseAppRef()

            else:
                f2d = None

            if self.fusion2d[0] is not None:
                a.deleteObjects(self.fusion2d[0])

            else:
                a.removeObjects(self.fusion2d[1], self.awindows)

            if len(fusobjs) == 0:
                self.fusion2d = []

            else:
                self.fusion2d = [f2d] + fusobjs

            # repalette( fusobjs )
            if f2d:
                obj = f2d

            elif len(fusobjs) == 1:
                obj = fusobjs[0]

            else:
                return

            self._displayVolume(obj, opts)

    def get_new_mesh2d_color(self):
        """Blabla"""

        colors = [
            (1.0, 0.3, 0.3, 1.0),
            (0.3, 1.0, 0.3, 1.0),
            (0.3, 0.3, 1.0, 1.0),
            (1.0, 1.0, 0.0, 1.0),
            (0.0, 1.0, 1.0, 1.0),
            (1.0, 0.0, 1.0, 1.0),
            (1.0, 1.0, 1.0, 1.0),
            (1.0, 0.7, 0.0, 1.0),
            (1.0, 0.0, 0.7, 1.0),
            (1.0, 0.7, 0.7, 1.0),
            (0.7, 1.0, 0.0, 1.0),
            (0.0, 1.0, 0.7, 1.0),
            (0.7, 1.0, 0.7, 1.0),
            (0.7, 0.0, 1.0, 1.0),
            (0.0, 0.7, 1.0, 1.0),
            (0.7, 0.7, 1.0, 1.0),
            (1.0, 1.0, 0.5, 1.0),
            (0.5, 1.0, 1.0, 1.0),
            (1, 0.5, 1.0, 1.0),
        ]
        used_cols = set([col for obj, col in six.itervalues(self.meshes2d)])
        for col in colors:
            if col not in used_cols:
                return col
        return len(self.meshes2d) % len(colors)

    def addMesh(self, obj, opts):
        """Blabla"""

        a = ana.Anatomist("-b")
        mesh2d = a.fusionObjects(
            [obj.getInternalRep()], method="Fusion2DMeshMethod"
        )
        color = self.get_new_mesh2d_color()
        self.meshes2d[obj.getInternalRep()] = (mesh2d, color)
        mesh2d.setMaterial(diffuse=color)
        mesh2d.releaseAppRef()
        windows_2d = [
            w
            for w in self.awindows
            if w.subtype()
            in (w.AXIAL_WINDOW, w.CORONAL_WINDOW, w.SAGITTAL_WINDOW)
        ]
        windows_3d = [w for w in self.awindows if w not in windows_2d]
        a.addObjects(mesh2d, windows_2d)
        a.addObjects(obj, windows_3d)

    def removeMesh(self, obj):
        """Blabla"""

        a = ana.Anatomist("-b")
        mesh2d, col = self.meshes2d[obj.getInternalRep()]
        a.removeObjects([obj, mesh2d], self.awindows)
        del self.meshes2d[obj.getInternalRep()]

    def disableButtons(self):
        """
        Disable plus or minus button depending on the selected object's display
        """
        self.setColorPalette()
        displayedObNames = []
        for i in range(len(self.displayedObjects)):
            displayedObNames.append(self.displayedObjects[i].name)
        item = Qt.QObject.findChild(
            self.awidget, QtCore.QObject, "objectslist"
        ).selectedItems()
        # There is always only one selected object
        if self.displayedObjects == [] or item == []:
            Qt.QObject.findChild(
                self.awidget, QtCore.QObject, "editAddAction"
            ).setEnabled(True)
        else:
            if item[0].text() in displayedObNames:
                Qt.QObject.findChild(
                    self.awidget, QtCore.QObject, "editAddAction"
                ).setEnabled(False)
                Qt.QObject.findChild(
                    self.awidget, QtCore.QObject, "editRemoveAction"
                ).setEnabled(True)
            else:
                Qt.QObject.findChild(
                    self.awidget, QtCore.QObject, "editRemoveAction"
                ).setEnabled(False)
                Qt.QObject.findChild(
                    self.awidget, QtCore.QObject, "editAddAction"
                ).setEnabled(True)

    def colorBackgroundList(self):
        """
        Color the background of displayed objects in objectlist and call
        changeIcon to add the right icon
        """
        displayedObNames = []
        for i in range(len(self.displayedObjects)):
            displayedObNames.append(self.displayedObjects[i].name)
        for i in range(len(self.aobjects)):
            item = Qt.QObject.findChild(
                self.awidget, QtCore.QObject, "objectslist"
            ).item(i)
            if item.text() in displayedObNames:
                self.changeIcon(item, i, "check")
                item = Qt.QObject.findChild(
                    self.awidget, QtCore.QObject, "objectslist"
                ).item(i)
                item.setBackground(QColor("#7fc97f"))
            else:
                self.changeIcon(item, i)
                item = Qt.QObject.findChild(
                    self.awidget, QtCore.QObject, "objectslist"
                ).item(i)
                item.setBackground(QColor("transparent"))

    def changeIcon(self, item, i, icon=None):
        """
        Adds empty icon if object is not displayed and check icon if displayed.
        """

        objectlist = Qt.QObject.findChild(
            self.awidget, QtCore.QObject, "objectslist"
        )
        row = objectlist.row(item)
        # remove item from objectlist
        objectlist.takeItem(row)
        object_name = self.aobjects[i].name
        # Add blank icon as spaceItem
        sources_images_dir = Config().getSourceImageDir()
        if icon == "check":
            icon = QIcon(os.path.join(sources_images_dir, "check.png"))
        else:
            icon = QIcon(os.path.join(sources_images_dir, "BLANK_ICON.png"))
        new_item = Qt.QListWidgetItem(icon, object_name)
        # reinsert new item with blank icon
        objectlist.insertItem(row, new_item)

    def addObject(self, obj):
        """
        Display an object in all windows
        """
        a = ana.Anatomist("-b")
        if obj not in self.displayedObjects:
            self.displayedObjects.append(obj)
        self.disableButtons()
        opts = {}
        if obj.objectType == "VOLUME":
            # volumes have a specific function since several volumes have to be
            # fusionned, and a volume rendering may occur
            self.addVolume(obj, opts)
            return
        elif obj.objectType == "SURFACE":
            self.addMesh(obj, opts)
            return
        elif obj.objectType == "GRAPH":
            opts["add_graph_nodes"] = 1

        a.addObjects(obj, self.awindows, **opts)

    def removeObject(self, obj):
        """
        Hides an object from views
        """
        a = ana.Anatomist("-b")
        if obj in self.displayedObjects:
            self.displayedObjects.remove(obj)
        self.disableButtons()
        if obj.objectType == "VOLUME":
            self.removeVolume(obj)
        elif obj.objectType == "SURFACE":
            self.removeMesh(obj)
        else:
            a.removeObjects(obj, self.awindows, remove_children=True)

    def fileOpen(self):
        """
        File browser + load object(s)
        """
        if not self.fdialog:
            self.fdialog = Qt.QFileDialog()
            self.fdialog.setDirectory(os.path.expanduser("~"))
        else:
            fd2 = self.fdialog
            self.fdialog = Qt.QFileDialog()
            self.fdialog.setDirectory(fd2.directory())
            self.fdialog.setHistory(fd2.history())
        self.fdialog.setFileMode(self.fdialog.ExistingFiles)
        self.fdialog.show()
        res = self.fdialog.exec_()
        if res:
            fnames = self.fdialog.selectedFiles()
            files = []
            for fname in fnames:
                print(six.text_type(fname))
                files.append(six.text_type(fname))
            self.loadObject(files)

    def selectedObjects(self):
        """
        list of objects selected in the list box on the upper left panel
        """
        olist = Qt.QObject.findChild(
            self.awidget, QtCore.QObject, "objectslist"
        )
        sobjs = []
        for o in olist.selectedItems():
            sobjs.append(six.text_type(o.text()).strip("\0"))
        return [o for o in self.aobjects if o.name in sobjs]

    def editAdd(self):
        """
        Display selected objects"""
        objs = self.selectedObjects()
        for o in objs:
            self.addObject(o)
        self.colorBackgroundList()

    def editRemove(self):
        """
        Hide selected objects"""
        objs = self.selectedObjects()
        for o in objs:
            self.removeObject(o)
        self.colorBackgroundList()

    def editDelete(self):
        """
        Delete selected objects"""
        objs = self.selectedObjects()
        self.deleteObjects(objs)

    def deleteObjects(self, objs):
        """Delete the given objects"""
        a = ana.Anatomist("-b")
        for o in objs:
            self.removeObject(o)
        olist = Qt.QObject.findChild(
            self.awidget, QtCore.QObject, "objectslist"
        )
        for o in objs:
            olist.takeItem(
                olist.row(olist.findItems(o.name, QtCore.Qt.MatchExactly)[0])
            )
        self.aobjects = [o for o in self.aobjects if o not in objs]
        a.deleteObjects(objs)

    def deleteObjectsFromFiles(self, files):
        """
        Delete the given objects given by their file names
        """
        a = ana.Anatomist("-b")
        objs = [o for o in a.getObjects() if o.filename in files]
        self.deleteObjects(objs)

    def closeAll(self, close_ana=True):
        """Exit"""

        print("Exiting Ana2")
        self.newWindow.close()
        a = ana.Anatomist("-b")
        # remove windows from their parent to prevent them to be brutally
        # deleted by Qt.
        w = None

        for w in self.awindows:
            try:
                w.hide()

            except Exception:
                continue  # window closed by Qt ?

            self.viewgridlay.removeWidget(w.internalRep._get())
            w.setParent(None)

        del w
        self.awindows = []
        self.displayedObjects = []
        self.viewgridlay = None
        self.volrender = None
        self.fusion2d = []
        self.mesh3d = {}
        self.aobjects = []
        self.awidget.close()
        self.awidget = None
        del self.fdialog

        if close_ana:
            a = ana.Anatomist()
            a.close()

    def stopVolumeRendering(self):
        """Disable volume rendering: show a slice instead"""

        a = ana.Anatomist("-b")

        if not self.volrender:
            return

        a.deleteObjects(self.volrender)
        self.volrender = None

        if len(self.fusion2d) != 0:
            if self.fusion2d[0] is not None:
                obj = self.fusion2d[0]

            else:
                obj = self.fusion2d[1]

        wins = [w for w in self.awindows if w.subtype() == 0]
        a.addObjects(obj, wins)
        self.control_3d_type = "LeftSimple3DControl"
        a.execute("SetControl", windows=wins, control=self.control_3d_type)

    def startVolumeRendering(self):
        """Enable volume rendering in 3D views"""
        a = ana.Anatomist("-b")
        if len(self.fusion2d) == 0:
            return
        if self.fusion2d[0] is not None:
            obj = self.fusion2d[0]
        else:
            obj = self.fusion2d[1]
        wins = [x for x in self.awindows if x.subtype() == 0]
        if len(wins) == 0:
            return
        vr = a.fusionObjects([obj], method="VolumeRenderingFusionMethod")
        vr.releaseAppRef()
        clip = a.fusionObjects([vr], method="FusionClipMethod")
        clip.releaseAppRef()
        self.volrender = [clip, vr]
        a.removeObjects(obj, wins)
        a.addObjects(clip, wins)
        self.control_3d_type = "VolRenderControl"
        a.execute("SetControl", windows=wins, control=self.control_3d_type)

    def enableVolumeRendering(self, on):
        """Enable/disable volume rendering in 3D views"""
        self._vrenabled = on
        if self._vrenabled:
            self.startVolumeRendering()
        else:
            self.stopVolumeRendering()

    def open_anatomist_main_window(self):
        """Blabla"""

        a = ana.Anatomist()
        cw = a.getControlWindow()
        a.execute("CreateControlWindow")
        if not cw:
            anacontrolmenu = sys.modules.get("anacontrolmenu")
            if anacontrolmenu:
                anacontrolmenu.add_gui_menus()

    def coordsChanged(self):
        """set the cursor on the position entered in the coords fields"""
        a = ana.Anatomist("-b")
        if len(self.awindows) == 0:
            return

        def findChild(x, y):
            """Blabla"""

            return Qt.QObject.findChild(x, QtCore.QObject, y)

        pos = [
            float(findChild(self.awidget, "coordXEdit").text()),
            float(findChild(self.awidget, "coordYEdit").text()),
            float(findChild(self.awidget, "coordZEdit").text()),
        ]
        # take coords transformation into account
        tr = a.getTransformation(
            a.mniTemplateRef, self.awindows[0].getReferential()
        )
        if tr is not None:
            pos = tr.transform(pos)
        t = float(findChild(self.awidget, "coordTEdit").text())
        a.execute(
            "LinkedCursor", window=self.awindows[0], position=pos[:3] + [t]
        )

    def dragEnterEvent(self, win, event):
        """Blabla"""

        x = ana.cpp.QAObjectDrag.canDecode(
            event
        ) or ana.cpp.QAObjectDrag.canDecodeURI(event)
        if x:
            event.accept()
        else:
            event.reject()

    def dropEvent(self, win, event):
        """Blabla"""

        a = ana.Anatomist("-b")
        o = ana.cpp.set_AObjectPtr()
        if ana.cpp.QAObjectDrag.decode(event, o):
            for obj in o:
                ob = a.AObject(o)
                if ob not in self.aobjects:
                    self.registerObject(ob)
                else:
                    self.addObject(ob)
            event.accept()
            return
        else:
            things = ana.cpp.QAObjectDrag.decodeURI(event)
            if things is not None:
                for obj in things[0]:
                    objnames = [x.fileName() for x in self.aobjects]
                    if obj not in objnames:
                        self.loadObject(obj)
                    else:
                        o = [x for x in self.aobjects if x.fileName() == obj][
                            0
                        ]
                        self.addObject(o)
                # TODO: things[1]: .ana scripts
                event.accept()
                return
        event.reject()

    def popup_objects(self):
        """
        Right-click popup on objects list
        """
        sel = self.selectedObjects()
        if len(sel) == 0:
            return
        t = aims.Tree()
        osel = [o.getInternalRep() for o in sel]
        menu = ana.cpp.OptionMatcher.popupMenu(osel, t)
        prop = menu.addAction("Object properties")
        prop.triggered.connect(self.object_properties)
        new_view = menu.addAction("Open in new view")
        new_view.triggered.connect(lambda: self.addNewView(sel[0]))
        menu.exec_(Qt.QCursor.pos())

    def object_properties(self):
        """
        Display selected objects properties in a browser window
        """
        a = ana.Anatomist("-b")
        if (
            not hasattr(self, "browser")
            or not self.browser
            or self.browser.isNull()
            or not self.browser.isVisible()
        ):
            self.browser = a.createWindow("Browser")
        else:
            self.browser.removeObjects(self.browser.Objects())
        self.browser.addObjects(self.selectedObjects())

    def addNewView(self, object):
        """
        Opens a popup with a view of the object
        Default display is axial view but can be changed
        """
        self.newWindow.showPopup(object)



# -*- coding: utf-8 -*-
"""This module is dedicated to pipeline history."""

import os.path as osp

import traits.api as traits
from capsul.api import Pipeline, Process  # , capsul_engine

from populse_mia.data_manager.project import (
    BRICK_EXEC,
    BRICK_EXEC_TIME,
    BRICK_ID,
    BRICK_INPUTS,
    BRICK_NAME,
    BRICK_OUTPUTS,
    COLLECTION_BRICK,
    COLLECTION_CURRENT,
    TAG_BRICKS,
)


class ProtoProcess(object):
    """
    Lightweight convenience class, stores a brick database entry, plus
    additional info (used)
    """

    def __init__(self, brick=None):
        self.brick = brick
        self.used = False


def data_history_pipeline(filename, project):
    """
    Get the complete "useful" history of a file in the database, as a "fake
    pipeline".

    The pipeline contains fake processes (unspecialized, direct Process
    instances), with all parameters (all being of type Any). The pipeline has
    connections, and gets all upstream ancestors of the file, so it contains
    all processing used to produce the latest version of the file (it may have
    been modified several time during the processing), and gets as inputs all
    input files which were used to produce the final data.

    Processing bricks which are not used, probably part of earlier runs which
    have been orphaned because the data file has been overwritten, are not
    listed in this history.
    """

    procs, links = get_data_history_processes(filename, project)

    if procs:
        pipeline = Pipeline()
        for proc in procs.values():
            if proc.used:
                pproc = brick_to_process(proc.brick, project)
                proc.process = pproc
                name = pproc.name
                if name in pipeline.nodes:
                    name = "%s_%s" % (name, pproc.uuid.replace("-", "_"))
                pproc.node_name = name
                pipeline.add_process(name, pproc)

        for link in links:
            if link[0] is None:
                src = link[1]
                if src not in pipeline.traits():
                    pipeline.export_parameter(
                        link[2].process.node_name, link[3], src
                    )
                    src = None
                elif pipeline.trait(src).output:
                    # already taken as an output: export under another name
                    done = False
                    n = 0
                    while not done:
                        src2 = "%s_%d" % (src, n)
                        if src2 not in pipeline.traits():
                            pipeline.export_parameter(
                                link[2].process.node_name, link[3], src2
                            )
                            src = None
                            done = True
                        elif not pipeline.trait(src2).output:
                            src = src2
                            done = True
                        n += 1
            else:
                src = "%s.%s" % (link[0].process.node_name, link[1])
            if link[2] is None:
                dst = link[3]
                if dst not in pipeline.traits():
                    pipeline.export_parameter(
                        link[0].process.node_name, link[1], dst
                    )
                    dst = None
                elif not pipeline.trait(dst).output:
                    # already taken as an input: export under another name
                    done = False
                    n = 0
                    while not done:
                        dst2 = "%s_%d" % (dst, n)
                        if dst2 not in pipeline.traits():
                            pipeline.export_parameter(
                                link[0].process.node_name, link[1], dst2
                            )
                            dst = None
                            done = True
                        elif pipeline.trait(dst2).output:
                            dst = dst2
                            done = True
                        n += 1
            else:
                dst = "%s.%s" % (link[2].process.node_name, link[3])
            if src is not None and dst is not None:
                try:
                    pipeline.add_link("%s->%s" % (src, dst))
                except ValueError as e:
                    print(e)

        return pipeline

    else:
        return None


def get_data_history_bricks(filename, project):
    """
    Get the complete "useful" history of a file in the database, as a set of
    bricks.

    This is just a fileterd version of :func:`get_data_history_processes` (like
    :func:`data_history_pipeline` in another shape), which only returns the set
    of brick elements actually used in the "useful" history of the data.
    """

    procs, links = get_data_history_processes(filename, project)
    bricks = {proc.brick for proc in procs.values() if proc.used}
    return bricks


def get_data_history(filename, project):
    """
    Get the processing history for the given data file. Based on
    :func:`get_data_history_processes`.

    The history dict contains several elements:

    - `parent_files`: set of other data used (directly or indirectly) to
      produce the data.
    - `processes`: processing bricks set from each ancestor data which
      lead to the given one. Elements are process (brick) UUIDs.

    Returns
    -------
    history: dict
    """
    procs, links = get_data_history_processes(filename, project)

    # parse input files

    parent_files = set()
    for proc in procs.values():
        if not proc.used:
            continue
        for value in proc.brick[BRICK_INPUTS].values():
            filenames = get_filenames_in_value(
                value, project, allow_temp=False
            )
            parent_files.update(filenames)

    bricks = {proc.brick[BRICK_ID] for proc in procs.values() if proc.used}
    history = {"processes": bricks, "parent_files": parent_files}

    return history


def get_data_history_processes(filename, project):
    """
    Get the complete "useful" history of a file in the database.

    The function outputs a dict of processes (:class:`ProtoProcess` instances)
    and a set of links between them. :func:`data_history_pipeline` is a
    higher-level function which is using this one, then converts its outputs
    into a :class:`~capsul.pipeline.pipeline.Pipeline` instance to represent
    the data history.

    The processes output by this function may include extra processes that are
    looked during history search, but finally not used. They are not filtered
    out, but they are distinguished with their ``used`` attribute: those
    actually used have it set to True.

    Processing bricks which are not used, probably part of earlier runs which
    have been orphaned because the data file has been overwritten, are either
    not listed in this history, or have their ``used`` property set to False.

    Returns
    -------
    procs: dict
        {uuid: ProtoProcess instance}
    links: set
        {(src_protoprocess, src_plug_name, dst_protoprocess, dst_plug_name)}.
        Links from/to the 'exterior" are also given: in this
        case src_protoprocess or dst_protoprocess is None.
    """

    # session = project.session

    procs = {}
    links = set()
    new_procs = get_direct_proc_ancestors(filename, project, procs)
    done_procs = set()

    # keep only the latest to begin with
    later_date = None
    keep_procs = {}
    for uuid, proc in new_procs.items():
        date = proc.brick[BRICK_EXEC_TIME]
        if later_date is None:
            later_date = date
            keep_procs[uuid] = proc
        elif date > later_date:
            later_date = date
            keep_procs = {uuid: proc}
        elif date == later_date:
            # ambiguity: keep all equivalent
            keep_procs[uuid] = proc
        else:
            print("drop earlier run:", proc.brick[BRICK_NAME], uuid)

    todo = list(keep_procs.values())

    while todo:
        proc = todo.pop(0)
        if proc in done_procs:
            continue
        done_procs.add(proc)
        proc.used = True

        print(
            "-- ancestors for:",
            proc.brick[BRICK_ID],
            proc.brick[BRICK_NAME],
            proc.brick[BRICK_EXEC_TIME],
        )
        values_w_files = {}
        for name, value in proc.brick[BRICK_INPUTS].items():
            filenames = get_filenames_in_value(value, project)
            # record inputs referencing files in the DB
            if filenames:
                print(name, "will be parsed.")
                values_w_files[name] = (value, filenames)

        for name, (value, filenames) in values_w_files.items():
            for nfilename in filenames:
                if nfilename == "<temp>":
                    print("temp file used -- history is broken")
                    prev_procs, prev_links = get_proc_ancestors_via_tmp(
                        proc, project, procs
                    )
                    links.update(prev_links)

                    n_procs = [
                        pproc
                        for pproc in prev_procs.values()
                        if pproc not in done_procs
                    ]
                    todo += n_procs
                else:
                    prev_procs = get_direct_proc_ancestors(
                        nfilename,
                        project,
                        procs,
                        before_exec_time=proc.brick[BRICK_EXEC_TIME],
                        org_proc=proc,
                    )

                    n_procs = [
                        pproc
                        for pproc in prev_procs.values()
                        if pproc not in done_procs
                    ]
                    todo += n_procs

                    # connect outputs of prev_procs which are identical to
                    print("look for value", value, "in", prev_procs.keys())
                    for pproc in prev_procs.values():
                        print("- in", pproc.brick[BRICK_NAME])
                        for pname, pval in pproc.brick[BRICK_OUTPUTS].items():
                            if pval == value or data_in_value(
                                pval, nfilename, project
                            ):
                                links.add((pproc, pname, proc, name))

                if len(prev_procs) == 0 or prev_procs == {
                    proc.brick[BRICK_ID]: proc
                }:
                    # the param has no previous processing or just the current
                    # self-modifing process: connect it to main inputs
                    links.add((None, name, proc, name))

    for proc in keep_procs.values():
        for name, value in proc.brick[BRICK_OUTPUTS].items():
            if data_in_value(value, filename, project):
                links.add((proc, name, None, name))

    print(
        "history of:",
        filename,
        ":",
        len([p for p in procs.values() if p.used]),
        "processes, ",
        len(links),
        "links",
    )
    return procs, links


def get_direct_proc_ancestors(
    filename,
    project,
    procs,
    before_exec_time=None,
    only_latest=True,
    org_proc=None,
):
    """
    Retrieve processing bricks which are referenced in the direct filename
    history. It can get the latest before a given execution time. As exec time
    is ambiguous (several processes may have finished at exactly the same
    time), several processes may be kept for a given exec time.

    The "origin" process, if given, is excluded from this exec time filtering
    (as we are looking for the one preceding it), but still included in the
    ancestors list.

    This function manipulates processing bricks as :class:`ProtoProcess`
    instances, a light wrapper for a brick database entry.

    Parameters
    ----------
    filename: str
        data filename to inspect
    project: Project instance
        used to access database
    procs: dict
        process dict, {uuid: ProtoProcess instance}, the dict is populated when
        bricks are retrieved from the database.
    before_exec_time: datetime instance (optional)
        if it is specified, only processing bricks not newer than this time are
        used.
    only_latest: bool (optional, default: True)
        if True, only the latest processes retrieved from the history are kept.
        If ``before_exec_time`` is also used, then it is the latest before this
        time.
    org_proc: ProtoProcess instance (optional)
        if filename is the output of a process, we can specify it here in order
        to exclude it from the time filtering (otherwise only this process will
        likely remain).

    Returns
    -------
    procs: dict
        {brick uuid: ProtoProcess instance}
    """

    session = project.session
    bricks = session.get_value(COLLECTION_CURRENT, filename, TAG_BRICKS)
    print("bricks for:", filename, ":", bricks)

    new_procs = {}
    # new_links = set()

    if bricks is not None:
        for brick in bricks:
            if brick not in procs:
                proc = get_history_brick_process(
                    brick, project, before_exec_time=before_exec_time
                )
                if proc is None:
                    continue

                procs[brick] = proc
                new_procs[brick] = proc
            else:
                proc = procs[brick]
                if (
                    before_exec_time
                    and proc.brick[BRICK_EXEC_TIME] > before_exec_time
                ):
                    continue
                new_procs[brick] = procs[brick]

    if only_latest:
        # keep last run(s)
        later_date = None
        keep_procs = {}
        for uuid, proc in new_procs.items():
            if org_proc and proc is org_proc:
                # ignore origin proc for date sorting
                continue
            date = proc.brick[BRICK_EXEC_TIME]
            if later_date is None:
                later_date = date
                keep_procs[uuid] = proc
            elif date > later_date:
                later_date = date
                keep_procs = {uuid: proc}
            elif date == later_date:
                # ambiguity: keep all equivalent
                keep_procs[uuid] = proc
            else:
                print("drop earlier run:", proc.brick[BRICK_NAME])
        if org_proc and org_proc.brick[BRICK_ID] in new_procs:
            # set back origin process, if it's in the list
            keep_procs[org_proc.brick[BRICK_ID]] = org_proc
    else:
        keep_procs = new_procs

    return keep_procs


def get_proc_ancestors_via_tmp(proc, project, procs):
    """
    Normally an internal function used in :func:`get_data_history_processes`
    and :func:`data_history_pipeline`: it is not meant to be part of a public
    API.

    Try to get upstream process(es) for proc, connected via a temp value
    ("<temp>").

    For this, try to match processes in the output files history bricks.

    Bricks are looked for, first in the process input files direct histories.
    If no matching process is found, then the full database bricks history is
    searched, which may be much slower for large databases.

    The matching is made by the "<temp>" filename and processing time, thus
    is error-prone, especially if searching the whole bricks database.

    ``proc`` should be a :class:`ProtoProcess` instance

    Returns
    -------
    new_procs: dict
        {uuid: ProtoProcess instance}
    links: set
        {(src_protoprocess, src_plug_name, dst_protoprocess, dst_plug_name)}.
        Pipeline link from/to the pipeline main plugs are also given: in this
        case src_protoprocess or dst_protoprocess is None.
    """

    new_procs = {}
    links = set()
    dlink = None
    tmp_filename = "<temp>"

    def _get_tmp_param(proc):
        """Blabla"""

        for param, value in proc.brick[BRICK_INPUTS].items():
            if data_in_value(value, tmp_filename, project):
                return (proc, param)
        return (None, None)  # failed...

    # look first from proc outputs history (which is more direct, less error-
    # prone, and a more limited search)
    for name, value in proc.brick[BRICK_OUTPUTS].items():
        filenames = get_filenames_in_value(value, project, allow_temp=False)
        for filename in filenames:
            hprocs = get_direct_proc_ancestors(
                filename,
                project,
                procs,
                before_exec_time=proc.brick[BRICK_EXEC_TIME],
                only_latest=False,
            )
            if proc.brick[BRICK_ID] in hprocs:
                # exclude the current proc
                del hprocs[proc.brick[BRICK_ID]]
            sprocs = find_procs_with_output(
                hprocs.values(), tmp_filename, project
            )
            for exec_time in sorted(sprocs, reverse=True):
                for hproc, param in sprocs[exec_time]:
                    new_procs[hproc.brick[BRICK_ID]] = hproc
                    if dlink is None:
                        dlink = _get_tmp_param(proc)
                    links.add((hproc, param, dlink[0], dlink[1]))
                    # we have found a link (starting with the older): stop
                    break
                if len(new_procs) != 0:
                    break
            # if found, should we still process other filenames ?

    if len(new_procs) == 0:
        # not found in data history: search the entire bricks histories
        session = project.session
        print("temp history not found from output filenames...")

        # print('test bricks older than:', proc.exec_time)
        # filtering for date <= doesn't seem to work as I expect...
        # bricks = session.filter_documents(
        # COLLECTION_BRICK, '{%s} <= "%s"' % (BRICK_EXEC_TIME, proc.exec_time))
        candidates = {}
        bricks = session.get_documents(COLLECTION_BRICK)
        for brick in bricks:
            # if brick
            if brick[BRICK_EXEC] != "Done":
                continue
            if brick[BRICK_EXEC_TIME] > proc.brick[BRICK_EXEC_TIME]:
                continue
            # print('try brick:', brick[BRICK_NAME])
            outputs = brick[BRICK_OUTPUTS]
            for name, value in outputs.items():
                if data_in_value(value, tmp_filename, project):
                    candidates.setdefault(brick[BRICK_EXEC_TIME], []).append(
                        (brick, name)
                    )
                    # print('CANDIDATE.')
                    break
        for exec_time in sorted(candidates, reverse=True):
            for brick, name in candidates[exec_time]:
                brick_id = brick[BRICK_ID]
                hproc = procs.get(brick_id)
                if hproc is None:
                    hproc = get_history_brick_process(brick_id, project)
                    procs[brick_id] = hproc
                new_procs[brick_id] = hproc
                if dlink is None:
                    dlink = _get_tmp_param(proc)
                links.add((hproc, name, dlink[0], dlink[1]))
                print("found:", hproc.brick[BRICK_NAME], name)
                break
            break

    return new_procs, links


def find_procs_with_output(procs, filename, project):
    """
    Find in the given process list if the given filename is part of
    its outputs.

    Parameters
    ----------
    procs: iterable
        process in the list is a :class:`ProtoProcess` instance
    filename: str
        a file name
    project: :class:`~populse_mia.data_manager.project.Project` instance
        used only to get the database folder (base directory for data)

    Returns
    -------
    sprocs: dict
        exec_time: [(process, param_name), ...]
    """

    sprocs = {}
    for proc in procs:
        for name, value in proc.brick[BRICK_OUTPUTS].items():
            if data_in_value(value, filename, project):
                sprocs.setdefault(proc.brick[BRICK_EXEC_TIME], []).append(
                    (proc, name)
                )
    return sprocs


def data_in_value(value, filename, project):
    """
    Looks if the given filename is part of the given value. The value may ba a
    list, a tuple, or a dict, and may include several layers, which are parsed.

    The input filename may be the temp value "<temp>", or a filename in its
    "short" version (relative path to the project database data directory).
    """

    if isinstance(value, str):
        if filename != "<temp>":
            proj_dir = osp.join(osp.abspath(osp.normpath(project.folder)), "")
            filename = osp.join(proj_dir, filename)
        return value == filename
    if isinstance(value, (list, tuple)):
        for val in value:
            if data_in_value(val, filename, project):
                return True
        return False
    if hasattr(value, "values"):
        for val in value.values():
            if data_in_value(val, filename, project):
                return True
    return False


def is_data_entry(filename, project, allow_temp=True):
    """
    Checks if the input filename is a database entry. The return value is
    either the relative path to the database data directory, or "<temp>" if
    filename is this value and allow_temp is True (which is the default), or
    None if it is not in the database.
    """

    if allow_temp and filename == "<temp>":
        return filename

    proj_dir = osp.join(osp.abspath(osp.normpath(project.folder)), "")

    if not filename.startswith(proj_dir):
        return None

    # fmt: off
    filename = filename[len(proj_dir):]
    # fmt: on

    if project.session.has_document(COLLECTION_CURRENT, filename):
        return filename

    return None


def get_filenames_in_value(value, project, allow_temp=True):
    """
    Parses ``value``, which may be an imbrication of lists, tuples and dicts,
    and gets all filenames referenced in it. Only filenames which are database
    entries are kept, and the "<temp>" value if ``allow_temp`` is True (which
    is the default). Other non-indexed filenames are considered to be read-only
    static data (such as templates, atlases or other software-related data),
    and are not retained.
    """

    values = [value]
    filenames = set()
    while values:
        value = values.pop(0)
        if isinstance(value, str):
            nvalue = is_data_entry(value, project, allow_temp=allow_temp)
            if nvalue:
                filenames.add(nvalue)
        elif isinstance(value, (list, tuple)):
            values.extend(value)
        elif hasattr(value, "values"):
            values.extend(value.values())

    return filenames


def get_history_brick_process(brick_id, project, before_exec_time=None):
    """
    Get a brick from its uuid in the database, and return it as a
    :class:`ProtoProcess` instance.

    A brick that has not been executed (its exec status is not ``"Done"``), or
    if it is newer than ``before_exec_time`` if this parameter is given, is
    discarded.

    If discarded (or nor found in the database), the return value is None.
    """

    session = project.session
    binfo = session.get_document(COLLECTION_BRICK, brick_id)
    if binfo is None:
        return None
    exec_status = binfo[BRICK_EXEC]
    if exec_status != "Done":
        return None
    exec_time = binfo[BRICK_EXEC_TIME]
    print(brick_id, "exec_time:", exec_time, ", before:", before_exec_time)
    if before_exec_time and exec_time > before_exec_time:
        # ignore later runs
        return None
    print(brick_id, ":", binfo[BRICK_NAME])

    proc = ProtoProcess(binfo)

    return proc


def brick_to_process(brick, project):
    """
    Converts a brick database entry (document) into a "fake process": a
    :class:`̀~capsul.process.process.Process` direct instance (not subclassed)
    which cannot do any actual processing, but which represents its parameters
    with values (traits and values). The process gets a ``name`` and an
    ``uuid`` from the brick, and also an ``exec_time``.
    """

    if isinstance(brick, str):
        # brick is an id
        session = project.session
        brick = session.get_document(COLLECTION_BRICK, brick)
    if brick is None:
        return None

    inputs = brick[BRICK_INPUTS]
    outputs = brick[BRICK_OUTPUTS]

    proc = Process()
    proc.name = brick[BRICK_NAME].split(".")[-1]
    proc.uuid = brick[BRICK_ID]
    proc.exec_time = brick[BRICK_EXEC_TIME]

    for name, value in inputs.items():
        proc.add_trait(name, traits.Any(output=False, optional=True))
        setattr(proc, name, value)

    for name, value in outputs.items():
        proc.add_trait(name, traits.Any(output=True, optional=True))
        setattr(proc, name, value)

    return proc



# -*- coding: utf-8 -*-
"""The first module used at the mia runtime.

Basically, this module is dedicated to the initialisation of the basic
parameters and the various checks necessary for a successful launch of the
mia's GUI.

:Contains:
    :Function:
        - main

"""

###############################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
###############################################################################

import os
import sys
import tempfile

# PyQt5 imports
from PyQt5.QtCore import QCoreApplication, Qt, qInstallMessageHandler
from PyQt5.QtWidgets import QApplication, QMessageBox

# main_window = None


def main():
    """Make basic configuration check, then actual launch of Mia.

    Checks if Mia is called from the site/dist packages (`user mode`) or from a
    cloned git repository (`developer mode`).

    ~/.populse_mia/configuration_path.yml is mandatory, if it doesn't exist
    or is corrupted, try to create one with a valid properties path.

    - If launched from a cloned git repository (`developer mode`):
        - the properties_path is the "properties_dev_path" parameter in
          ~/.populse_mia/configuration_path.yml
    - If launched from the site/dist packages (`user mode`):
        - the properties_path is the "properties_user_path" parameter in
          ~/.populse_mia/configuration_path.yml

    Launches the verify_processes() function, then the launch_mia() function
    (Mia's real launch).
    """

    pypath = []
    package_not_found = []

    # Disables any etelemetry check.
    if "NO_ET" not in os.environ:
        os.environ["NO_ET"] = "1"

    if "NIPYPE_NO_ET" not in os.environ:
        os.environ["NIPYPE_NO_ET"] = "1"

    # Trying to fix High DPI Display issue
    os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"

    # General QApplication class instantiation
    QCoreApplication.setAttribute(Qt.AA_ShareOpenGLContexts)
    app = QApplication(sys.argv)
    QApplication.setOverrideCursor(Qt.WaitCursor)

    # Adding the populse projects path to sys.path, if in developer mode
    if (
        not os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        in sys.path
    ):  # "developer" mode
        DEV_MODE = True
        os.environ["MIA_DEV_MODE"] = "1"
        root_dev_dir = os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
        )
        branch = ""
        populse_bdir = ""
        capsul_bdir = ""
        soma_bdir = ""

        if not os.path.isdir(os.path.join(root_dev_dir, "populse_mia")):
            # Different sources layout - try casa_distro mode
            root_dev_dir = os.path.dirname(
                os.path.dirname(
                    os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
                )
            )

            if os.path.basename(root_dev_dir) == "populse":
                root_dev_dir = os.path.dirname(root_dev_dir)
                populse_bdir = "populse"
                soma_bdir = "soma"

            print("root_dev_dir:", root_dev_dir)
            branch = os.path.basename(
                os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
            )
            print("branch:", branch)

        i = 0
        # Adding populse_mia
        print('\n- Mia in "developer" mode')
        mia_dev_dir = os.path.join(
            root_dev_dir, populse_bdir, "populse_mia", branch
        )
        print("  . Using populse_mia package from {} ...".format(mia_dev_dir))
        sys.path.insert(i, mia_dev_dir)
        pypath.append(mia_dev_dir)
        del mia_dev_dir
        from populse_mia import info

        print(f"    populse_mia version: {info.__version__}")

        # Adding capsul
        if os.path.isdir(os.path.join(root_dev_dir, capsul_bdir, "capsul")):
            i += 1
            capsul_dev_dir = os.path.join(
                root_dev_dir, capsul_bdir, "capsul", branch
            )
            print(
                "  . Using capsul package from {} ...".format(capsul_dev_dir)
            )
            sys.path.insert(i, capsul_dev_dir)
            pypath.append(capsul_dev_dir)
            del capsul_dev_dir

        else:

            try:
                import capsul

            except Exception:
                print("  . capsul package was not found!...")
                package_not_found.append("capsul")

            else:
                capsul_dir = os.path.dirname(os.path.dirname(capsul.__file__))
                print(
                    "  . Using capsul package from {} ...".format(capsul_dir)
                )
                del capsul_dir
                del capsul

        # Adding soma_base:
        if os.path.isdir(os.path.join(root_dev_dir, soma_bdir, "soma-base")):
            i += 1
            soma_b_dev_dir = os.path.join(
                root_dev_dir, soma_bdir, "soma-base", branch, "python"
            )
            print("  . Using soma package from {} ...".format(soma_b_dev_dir))
            sys.path.insert(i, soma_b_dev_dir)
            pypath.append(soma_b_dev_dir)
            del soma_b_dev_dir

        else:

            try:
                import soma

            except Exception:
                print("  . soma package was not found!...")
                package_not_found.append("soma")

            else:
                soma_b_dir = os.path.dirname(os.path.dirname(soma.__file__))
                print("  . Using soma package from {} ...".format(soma_b_dir))
                del soma_b_dir
                del soma

        # Adding soma_workflow:
        if os.path.isdir(
            os.path.join(root_dev_dir, soma_bdir, "soma-workflow")
        ):
            i += 1
            soma_w_dev_dir = os.path.join(
                root_dev_dir, soma_bdir, "soma-workflow", branch, "python"
            )
            print(
                "  . Using soma_workflow package from {} "
                "...".format(soma_w_dev_dir)
            )
            sys.path.insert(i, soma_w_dev_dir)
            pypath.append(soma_w_dev_dir)
            del soma_w_dev_dir

        else:

            try:
                import soma_workflow

            except Exception:
                print("  . soma_workflow package was not found!...")
                package_not_found.append("soma_workflow")

            else:
                soma_w_dir = os.path.dirname(
                    os.path.dirname(soma_workflow.__file__)
                )
                print(
                    "  . Using soma_worflow package from {} ...".format(
                        soma_w_dir
                    )
                )
                del soma_w_dir
                del soma_workflow

        # Adding populse_db:
        if os.path.isdir(
            os.path.join(root_dev_dir, populse_bdir, "populse_db")
        ):
            i += 1
            populse_db_dev_dir = os.path.join(
                root_dev_dir, populse_bdir, "populse_db", branch, "python"
            )
            print(
                "  . Using populse_db package from {} "
                "...".format(populse_db_dev_dir)
            )
            sys.path.insert(i, populse_db_dev_dir)
            pypath.append(populse_db_dev_dir)
            del populse_db_dev_dir

        else:

            try:
                import populse_db

            except Exception:
                print("  . populse_db package was not found!...")
                package_not_found.append("populse_db")

            else:
                populse_db_dir = os.path.dirname(
                    os.path.dirname(populse_db.__file__)
                )
                print(
                    "  . Using populse_db package from {} ...".format(
                        populse_db_dir
                    )
                )
                del populse_db_dir
                del populse_db

        # Adding mia_processes:
        if os.path.isdir(
            os.path.join(root_dev_dir, populse_bdir, "mia_processes")
        ):
            i += 1
            mia_processes_dev_dir = os.path.join(
                root_dev_dir, populse_bdir, "mia_processes", branch
            )
            print(
                "  . Using mia_processes package from {} "
                "...".format(mia_processes_dev_dir)
            )
            sys.path.insert(i, mia_processes_dev_dir)
            pypath.append(mia_processes_dev_dir)
            del mia_processes_dev_dir
            from mia_processes import info

            print(f"    mia_processes version: {info.__version__}")

        else:

            try:
                import mia_processes

            except Exception:
                print("  . mia_processes package was not found!...")
                package_not_found.append("mia_processes")

            else:
                mia_processes_dir = os.path.dirname(
                    os.path.dirname(mia_processes.__file__)
                )
                print(
                    "  . Using mia_processes package from {} "
                    "...".format(mia_processes_dir)
                )
                del mia_processes_dir
                del mia_processes

        del root_dev_dir

        if package_not_found:
            print(
                "\nMia cannot be started because the following packages "
                "were not found:\n"
                + "\n".join(f"- {package}" for package in package_not_found)
            )
            sys.exit(1)

    # elif "CASA_DISTRO" in os.environ:
    #     # If the casa distro development environment is detected,
    #     # developer mode is activated.
    #     os.environ["MIA_DEV_MODE"] = "1"
    #     DEV_MODE = True

    else:  # "user" mode
        os.environ["MIA_DEV_MODE"] = "0"
        DEV_MODE = False
        print('\n- Mia in "user" mode')
        # Where do we import the modules from?
        modules = (
            "populse_mia",
            "capsul",
            "soma",
            "soma_workflow",
            "populse_db",
            "mia_processes",
        )

        for i in modules:
            if i in sys.modules:
                mod = sys.modules[i]

            else:
                mod = __import__(i)
                del sys.modules[i]

            print(
                "  . Using {0} package from {1} ...".format(
                    mod.__name__, mod.__path__[0]
                )
            )

    # Check if nipype, mia_processes and capsul are available on the station.
    # If not available ask the user to install them
    pkg_error = []
    # pkg_error: a list containing nipype and/or mia_processes and/or capsul,
    #            if not currently installed
    capsulVer = None  # capsul version currently installed
    miaProcVer = None  # mia_processes version currently installed
    nipypeVer = None  # nipype version currently installed

    try:
        from capsul import info as capsul_info

        capsulVer = capsul_info.__version__

    except (ImportError, AttributeError) as e:
        pkg_error.append("capsul")
        print("\n" + "*" * 37)
        print("MIA warning {0}: {1}".format(e.__class__, e))
        print("*" * 37 + "\n")

    try:
        __import__("nipype")
        nipypeVer = sys.modules["nipype"].__version__

    except (ImportError, AttributeError) as e:
        pkg_error.append("nipype")
        print("\n" + "*" * 37)
        print("MIA warning {0}: {1}".format(e.__class__, e))
        print("*" * 37 + "\n")

    try:
        __import__("mia_processes")
        miaProcVer = sys.modules["mia_processes"].__version__

    except (ImportError, AttributeError) as e:
        pkg_error.append("mia_processes")
        print("\n" + "*" * 37)
        print("MIA warning {0}: {1}".format(e.__class__, e))
        print("*" * 37 + "\n")

    if pkg_error:
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Warning)
        msg.setWindowTitle("populse_mia -  warning: ImportError!")

        if len(pkg_error) == 1:
            msg.setText(
                "An issue has been detected with the {0} package. "
                "Please (re)install this package and/or fix the "
                "problems displayed in the standard output. "
                "Then, start again Mia ...".format(pkg_error[0])
            )

        elif len(pkg_error) == 2:
            msg.setText(
                "An issue has been detected with the {0} and {1} packages. "
                "Please (re)install these package and/or fix the "
                "problems displayed in the standard output. "
                "Then, start again Mia ...".format(pkg_error[0], pkg_error[1])
            )

        else:
            msg.setText(
                "An issue has been detected with the {0}, {1} and {2} "
                "packages. Please (re)install these package and/or fix the "
                "problems displayed in the standard output. "
                "Then, start again Mia ...".format(
                    pkg_error[0], pkg_error[1], pkg_error[2]
                )
            )

        msg.setStandardButtons(QMessageBox.Ok)
        msg.buttonClicked.connect(msg.close)
        msg.exec()
        sys.exit(1)

    # Now that populse projects paths have been set in sys.path, if necessary,
    # we can import from these projects:
    # Populse_mia imports
    from populse_mia.utils import (  # noqa E402
        check_python_version,
        launch_mia,
        verify_processes,
        verify_setup,
    )

    verify_setup(dev_mode=DEV_MODE, pypath=pypath)
    verify_processes(nipypeVer, miaProcVer, capsulVer)
    check_python_version()
    cwd = os.getcwd()

    with tempfile.TemporaryDirectory() as temp_work_dir:
        os.chdir(temp_work_dir)
        launch_mia(app)
        os.chdir(cwd)


def qt_message_handler(mode, context, message):
    """Custom Qt message handler to filter out specific messages"""
    for unwanted_message in unwanted_messages:

        if message.strip() == unwanted_message:
            return

        elif unwanted_message in message:
            # Remove the unwanted message but keep the rest of the line
            message = message.replace(unwanted_message, "").strip()

    # Output the remaining message (if any)
    if message:
        sys.stderr.write(message + "\n")


if __name__ == "__main__":
    # this will only be executed when this module is run directly
    # list of unwanted messages to filter out in stdout
    unwanted_messages = [
        "QPixmap::scaleHeight: Pixmap is a null pixmap",
    ]

    # Install the custom Qt message handler
    qInstallMessageHandler(qt_message_handler)
    main()



# -*- coding: utf-8 -*-
"""Define software version, description and requirements

:Contains:
    :Function:
        - get_populse_mia_gitversion

"""

###############################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
###############################################################################

import os
import subprocess
import sys

# Current version
version_major = 2
version_minor = 6
version_micro = 0
# version_extra = "dev"  # leave empty for release
version_extra = ""

# Expected by setup.py: string of form "X.Y.Z"
if version_extra:
    __version__ = "{0}.{1}.{2}-{3}".format(
        version_major, version_minor, version_micro, version_extra
    )

else:
    __version__ = "{0}.{1}.{2}".format(
        version_major, version_minor, version_micro
    )


def get_populse_mia_gitversion():
    """Mia version as reported by the last commit in git

    :returns: the version or None if nothing was found
    """

    try:
        import populse_mia

        dir_mia = os.path.realpath(
            os.path.join(
                os.path.dirname(populse_mia.__file__),
                os.path.pardir,
            )
        )

    except Exception:
        dir_mia = os.getcwd()

    dir_miagit = os.path.join(dir_mia, ".git")

    if not os.path.exists(dir_miagit):
        return None

    ver = None

    try:
        gitversion, _ = subprocess.Popen(
            "git show -s --format=%h",
            shell=True,
            cwd=dir_mia,
            stdout=subprocess.PIPE,
        ).communicate()

    except Exception:
        pass

    else:
        ver = gitversion.decode().strip().split("-")[-1]

    return ver


if __version__.endswith("-dev"):
    gitversion = get_populse_mia_gitversion()

    if gitversion:
        __version__ = "{0}+{1}".format(__version__, gitversion)

# Expected by setup.py: the status of the project
CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: CEA CNRS Inria "
    "Logiciel Libre License, version 2.1 (CeCILL-2.1)",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3 :: Only",
    "Topic :: Scientific/Engineering",
    "Topic :: Utilities",
]

# project descriptions
DESCRIPTION = "populse mia"
LONG_DESCRIPTION = """
===============
populse_mia
===============
[MIA] Multi parametric Image Analysis:
A complete image processing environment mainly targeted at
the analysis and visualization of large amounts of MRI data
"""

# Other values used in setup.py
NAME = "populse_mia"
ORGANISATION = "populse"
MAINTAINER = "Populse team"
MAINTAINER_EMAIL = "populse-support@univ-grenoble-alpes.fr"
AUTHOR = "Populse team"
AUTHOR_EMAIL = "populse-support@univ-grenoble-alpes.fr"
URL = "http://populse.github.io/populse_mia"
DOWNLOAD_URL = "http://populse.github.io/populse_mia"
LICENSE = "CeCILL"
VERSION = __version__
CLASSIFIERS = CLASSIFIERS
PLATFORMS = "OS Independent"

REQUIRES = [
    "capsul >= 2.5.0, < 3.0.0",
    "cryptography",
    "matplotlib",
    "mia-processes >= 2.5.0, < 3.0.0",
    "nibabel",
    "nipype",
    "pillow",
    "populse-db >= 2.5.0, < 3.0.0",
    "pre-commit",
    "pyqt5",
    "python-dateutil",
    "pyyaml",
    "scikit-image",
    "scipy",
    "snakeviz",
    "soma-base >= 5.2.0, < 6.0.0",
    "soma-workflow >= 3.2.2",
    "six >= 1.13",
    "traits",
]

EXTRA_REQUIRES = {
    "doc": [
        "sphinx>=1.0",
    ],
}

brainvisa_build_model = "pure_python"

# tests to run
test_commands = ["%s -m populse_mia.test" % sys.executable]



# -*- coding: utf-8 -*-
"""Module that handle the configuration of the software

Load and save the parameters from the miniviewer and the MIA preferences
pop-up in the config.yml file.

:Contains:
    :Class:
        - Config

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import glob
import os
import platform

import yaml

# Capsul import
from capsul.api import capsul_engine
from cryptography.fernet import Fernet

CONFIG = b"5YSmesxZ4ge9au2Bxe7XDiQ3U5VCdLeRdqimOOggKyc="


class Config:
    """
    Object that handles the configuration of the software

    :Contains:
        :Methods:
            - get_admin_hash: get the value of the hash of the admin password
            - get_afni_path: returns the path of AFNI
            - get_ants_path: returns the path of ANTS
            - getBackgroundColor: get background color
            - get_capsul_config: get CAPSUL config dictionary
            - get_capsul_engine: get a global CapsulEngine object used for all
              operations in MIA application
            - getChainCursors: returns if the "chain cursors" checkbox of the
              mini viewer is activated
            - get_freesurfer_setup: get freesurfer path
            - get_fsl_config: returns the path of the FSL config file
            - get_mainwindow_maximized: get the maximized (full-screen) flag
            - get_mainwindow_size: get the main window size
            - get_matlab_command: returns Matlab command
            - get_matlab_path: returns the path of Matlab's executable
            - get_matlab_standalone_path: returns the path of Matlab Compiler
              Runtime
            - get_max_projects: returns the maximum number of projects
              displayed in the "Saved projects" menu
            - get_max_thumbnails:  get max thumbnails number at the data
              browser bottom
            - get_properties_path: returns the software's properties path
            - get_mri_conv_path: returns the MRIManager.jar path
            - get_mrtrix_path: returns mrtrix path
            - getNbAllSlicesMax: returns the maximum number of slices to
              display in the mini viewer
            - get_opened_projects: returns the opened projects
            - getPathToProjectsFolder: returns the project's path
            - get_projects_save_path: returns the folder where the projects
              are saved
            - get_referential: returns boolean to indicate DataViewer
              referential
            - get_resources_path: get the resources path
            - getShowAllSlices: returns if the "show all slices" checkbox of
              the mini viewer is activated
            - getSourceImageDir: get the source directory for project images
            - get_spm_path: returns the path of SPM12 (license version)
            - get_spm_standalone_path: returns the path of SPM12 (standalone
              version)
            - getTextColor: return the text color
            - getThumbnailTag: returns the tag that is displayed in the mini
              viewer
            - get_use_afni: returns the value of "use afni" checkbox in the
              preferences
            - get_use_ants: returns the value of "use ants" checkbox in the
              preferences
            - get_use_clinical: returns the value of "clinical mode" checkbox
              in the preferences
            - get_use_freesurfer: returns the value of "use freesurfer"
              checkbox in the preferences
            - get_use_fsl: returns the value of "use fsl" checkbox in the
              preferences
            - get_use_matlab: returns the value of "use matlab" checkbox in the
              preferences
            - get_use_matlab_standalone: returns the value of "use matlab
              standalone" checkbox in the preferences
            - get_use_mrtrix: returns the value of "use mrtrix" checkbox in the
              prefrence
            - get_user_level: get the user level in the Capsul config
            - get_user_mode: returns the value of "user mode" checkbox
              in the preferences
            - get_use_spm: returns the value of "use spm" checkbox in the
              preferences
            - get_use_spm_standalone: returns the value of "use spm standalone"
              checkbox in the preferences
            - getViewerConfig: returns the DataViewer configuration (neuro or
              radio), by default neuro
            - getViewerFramerate: returns the DataViewer framerate for
              automatic time running images
            - isAutoSave: checks if auto-save mode is activated
            - isControlV1: checks if the selected display of the controller is
              of V1 type
            - isRadioView: checks if miniviewer in radiological orientation (if
              not, then it is in neurological orientation)
            - loadConfig: reads the config in the config.yml file
            - saveConfig: saves the config to the config.yml file
            - set_admin_hash: set the password hash
            - set_afni_path: set the path of the AFNI
            - set_ants_path: set the path of the ANTS
            - set_mrtrix_path: set the path of mrtrix
            - setAutoSave: sets the auto-save mode
            - setBackgroundColor: sets the background color
            - set_capsul_config: set CAPSUL configuration dict into MIA config
            - setChainCursors: set the "chain cursors" checkbox of the mini
              viewer
            - set_clinical_mode: set the value of "clinical mode" in
              the preferences
            - setControlV1: Set controller display mode (True if V1)
            - set_freesurfer_setup: set freesurfer path
            - set_fsl_config: set the path of the FSL config file
            - set_mainwindow_maximized: set the maximized (fullscreen) flag
            - set_mainwindow_size: set main window size
            - set_matlab_path: set the path of Matlab's executable
            - set_matlab_standalone_path: set the path of Matlab Compiler
              Runtime
            - set_max_projects: set the maximum number of projects displayed in
              the "Saved projects" menu
            - set_max_thumbnails: set max thumbnails number at the data browser
              bottom
            - set_mri_conv_path: set the MRIManager.jar path
            - setNbAllSlicesMax: set the maximum number of slices to display in
              the mini viewer
            - set_opened_projects: set the opened projects
            - set_projects_save_path: set the folder where the projects are
              saved
            - set_radioView: set the orientation in miniviewer (True for
              radiological, False for neurological orientation)
            - set_referential: set the DataViewer referential
            - set_resources_path: Set the resources path
            - setShowAllSlices: set the "show all slices" checkbox of the mini
              viewer
            - setSourceImageDir: set the source directory for project images
            - set_spm_path: set the path of SPM12 (license version)
            - set_spm_standalone_path: set the path of SPM12 (standalone
              version)
            - setTextColor: set the text color
            - setThumbnailTag: set the tag that is displayed in the mini viewer
            - set_use_afni: set the value of "use afni" checkbox in the
              preferences
            - set_use_ants: set the value of "use ants" checkbox in the
              preferences
            - set_use_freesurfer: set the value of "use freesurfer" checkbox
               in the preferences
            - set_use_fsl: set the value of "use fsl" checkbox in the
              preferences
            - set_use_matlab: set the value of "use matlab" checkbox in the
              preferences
            - set_use_matlab_standalone: set the value of "use matlab
              standalone" checkbox in the preferences
            - set_use_mrtrix: set the value of "use mrtrix" checkbox in the
              preferences
            - set_user_mode: set the value of "user mode" checkbox in
              the preferences
            - set_use_spm: set the value of "use spm" checkbox in the
              preferences
            - set_use_spm_standalone: set the value of "use spm standalone"
              checkbox in the preferences
            - setViewerConfig: set the Viewer configuration neuro or radio
            - setViewerFramerate: set the Viewer frame rate for automatic
              running time images
            - update_capsul_config: update a global CapsulEngine object used
              for all operations in MIA application
    """

    capsul_engine = None

    def __init__(self, properties_path=None):
        """Initialization of the Config class

        :Parameters:
            - :config_path: str (optional). If provided, the configuration file
               will be loaded/saved from the given directory. Otherwise, the
               regular heuristics will be used to determine the config path.
               This option allows to use an alternative config directory (for
               tests for instance).
        """

        if properties_path is not None:
            self.properties_path = properties_path

        if os.environ.get("MIA_DEV_MODE", None) is not None:
            self.dev_mode = bool(int(os.environ["MIA_DEV_MODE"]))

        else:
            # FIXME: What can we do if "MIA_DEV_MODE" is not in os.environ?
            print("\nMIA_DEV_MODE not found...\n")

        self.config = self.loadConfig()

    def get_admin_hash(self):
        """Get the value of the hash of the admin password.

        :returns: string
        """

        try:
            return self.config["admin_hash"]
        except KeyError:
            return False

    def get_afni_path(self):
        """Get the AFNI path

        :returns: string of path to AFNI
        """

        return self.config.get("afni", "")

    def get_ants_path(self):
        """Get the ANTS path

        :returns: string of path to ANTS
        """

        return self.config.get("ants", "")

    def get_freesurfer_setup(self):
        """Get the freesurfer path

        :returns: string of path to freesurfer
        """
        return self.config.get("freesurfer_setup", "")

    def getBackgroundColor(self):
        """Get background color.

        :returns: string of the background color
        """

        return self.config.get("background_color", "")

    def get_capsul_config(self, sync_from_engine=True):
        """Get CAPSUL config dictionary.

        :param sync_from_engine: if True, perform a synchronisation
                                 with CAPSUL engine (Bool)

        :return: capsul_config: the CAPSUL configuration saved into
                                Mia config (Dict)
        """

        capsul_config = self.config.setdefault("capsul_config", {})
        capsul_config.setdefault(
            "engine_modules",
            [
                "nipype",
                "fsl",
                "freesurfer",
                "matlab",
                "spm",
                "fom",
                "python",
                "afni",
                "ants",
                "mrtrix",
                "somaworkflow",
            ],
        )
        econf = capsul_config.setdefault("engine", {})
        eeconf = econf.setdefault("global", {})

        # Take mia pref parameters
        use_spm = self.get_use_spm()
        spm_path = self.get_spm_path()

        use_spm_standalone = self.get_use_spm_standalone()
        spm_standalone_path = self.get_spm_standalone_path()

        use_matlab = self.get_use_matlab()
        matlab_path = self.get_matlab_path()

        use_matlab_standalone = self.get_use_matlab_standalone()
        matlab_standalone_path = self.get_matlab_standalone_path()

        use_fsl = self.get_use_fsl()
        fsl_config = self.get_fsl_config()

        use_afni = self.get_use_afni()
        afni_path = self.get_afni_path()

        use_ants = self.get_use_ants()
        ants_path = self.get_ants_path()

        use_freesurfer = self.get_use_freesurfer()
        freesurfer_setup = self.get_freesurfer_setup()

        use_mrtrix = self.get_use_mrtrix()
        mrtrix_path = self.get_mrtrix_path()

        # Make synchronisation from Mia pref to Capsul config:

        # TODO: We do not deal with the version parameter. This can produce
        #       hidden configurations (spm and spm12 and spm8 ...)!

        # SPM
        spm_configs = eeconf.setdefault("capsul.engine.module.spm", {})
        if use_spm_standalone:
            m = {}
            for config_id, m in spm_configs.items():
                if (
                    m.get("standalone")
                    and m.get("directory") == spm_standalone_path
                ):
                    break
            if m:
                config_id = m["config_id"]
            else:
                config_id = "spm-standalone"
            m = spm_configs.setdefault(config_id, {})
            m.update(
                {
                    "config_id": config_id,
                    "config_environment": "global",
                    "directory": spm_standalone_path,
                    "standalone": True,
                }
            )

        if use_spm:
            m = {}
            for config_id, m in spm_configs.items():
                if not m.get("standalone") and m.get("directory") == spm_path:
                    break
            if m:
                config_id = m["config_id"]
            else:
                config_id = "spm"
            m = spm_configs.setdefault(config_id, {})
            m.update(
                {
                    "config_id": config_id,
                    "config_environment": "global",
                    "directory": spm_path,
                    "standalone": False,
                }
            )

        # MATLAB
        if use_matlab_standalone:
            m = eeconf.setdefault(
                "capsul.engine.module.matlab", {}
            ).setdefault("matlab", {})
            m["mcr_directory"] = matlab_standalone_path
            m["config_id"] = "matlab"
            m["config_environment"] = "global"

        if use_matlab:
            m = eeconf.setdefault(
                "capsul.engine.module.matlab", {}
            ).setdefault("matlab", {})
            m["executable"] = matlab_path
            m["config_id"] = "matlab"
            m["config_environment"] = "global"

        # FSL
        if use_fsl:
            m = eeconf.setdefault("capsul.engine.module.fsl", {}).setdefault(
                "fsl", {}
            )
            m["config_id"] = "fsl"
            m["config_environment"] = "global"
            m["config"] = fsl_config
            m["directory"] = os.path.dirname(
                os.path.dirname(os.path.dirname(fsl_config))
            )

        # AFNI
        if use_afni:
            m = eeconf.setdefault("capsul.engine.module.afni", {}).setdefault(
                "afni", {}
            )
            m["config_id"] = "afni"
            m["config_environment"] = "global"
            m["directory"] = afni_path

        # ANTS
        if use_ants:
            m = eeconf.setdefault("capsul.engine.module.ants", {}).setdefault(
                "ants", {}
            )
            m["config_id"] = "ants"
            m["config_environment"] = "global"
            m["directory"] = ants_path

        # Freesurfer
        if use_freesurfer:
            m = eeconf.setdefault(
                "capsul.engine.module.freesurfer", {}
            ).setdefault("freesurfer", {})
            m["config_id"] = "freesurfer"
            m["config_environment"] = "global"
            m["setup"] = freesurfer_setup
            # TODO: change fs subject dir
            m["subjects_dir"] = ""

        # mrtrix
        if use_mrtrix:
            m = eeconf.setdefault(
                "capsul.engine.module.mrtrix", {}
            ).setdefault("mrtrix", {})
            m["config_id"] = "mrtrix"
            m["config_environment"] = "global"
            m["directory"] = mrtrix_path

        # attributes completion
        m = eeconf.setdefault(
            "capsul.engine.module.attributes", {}
        ).setdefault("attributes", {})
        m["config_id"] = "attributes"
        m["config_environment"] = "global"
        m["attributes_schema_paths"] = [
            "capsul.attributes.completion_engine_factory",
            "populse_mia.user_interface.pipeline_manager.process_mia",
        ]
        m["process_completion"] = "mia_completion"

        # Synchronise from engine, or not
        if sync_from_engine and self.capsul_engine:
            for environment in (
                self.capsul_engine.settings.get_all_environments
            )():
                eeconf = econf.setdefault(environment, {})
                # would need a better merging system
                eeconf.update(
                    self.capsul_engine.settings.export_config_dict(
                        environment
                    )[environment]
                )

        return capsul_config

    @staticmethod
    def get_capsul_engine():
        """
        Get a global CapsulEngine object used for all operations in MIA
        application. The engine is created once when needed.

        :returns: Config.capsul_engine: capsul.engine.CapsulEngine object
        """

        config = Config()
        config.get_capsul_config()

        if Config.capsul_engine is None:
            Config.capsul_engine = capsul_engine()
            Config().update_capsul_config()

        return Config.capsul_engine

    def getChainCursors(self):
        """Get the value of the checkbox 'chain cursor' in miniviewer.

        :returns: boolean
        """

        return self.config.get("chain_cursors", False)

    def get_fsl_config(self):
        """Get the FSL config file  path

        :returns: string of path to the fsl/etc/fslconf/fsl.sh file
        """

        return self.config.get("fsl_config", "")

    def get_mainwindow_maximized(self):
        """Get the maximized (fullscreen) flag

        :returns:  boolean
        """

        return self.config.get("mainwindow_maximized", True)

    def get_mainwindow_size(self):
        """Get the main window size

        :returns:  list or None
        """

        return self.config.get("mainwindow_size", None)

    def get_matlab_command(self):
        """Get Matlab command.

        :returns: matlab executable path or nothing if matlab path not
                  specified
        """

        if self.config.get("use_spm_standalone"):
            archi = platform.architecture()
            if "Windows" in archi[1]:
                spm_script = glob.glob(
                    os.path.join(
                        self.config["spm_standalone"],
                        "spm*_win" + archi[0][:2] + ".exe",
                    )
                )
            else:
                spm_script = glob.glob(
                    os.path.join(self.config["spm_standalone"], "run_spm*.sh")
                )
            if spm_script:
                spm_script = spm_script[0]
                return "{0} {1} script".format(
                    spm_script, self.config["matlab_standalone"]
                )
            else:
                return ""  # will not work.
        else:
            return self.config.get("matlab", None)

    def get_matlab_path(self):
        """Get the path to the matlab executable.

        :returns: String of path
        """

        return self.config.get("matlab", None)

    def get_matlab_standalone_path(self):
        """Get the path to matlab compiler runtime.

        :returns: string of path
        """

        return self.config.get("matlab_standalone", "")

    def get_max_projects(self):
        """Get the maximum number of projects displayed in the "Saved
        projects" menu.

        :returns: Integer
        """

        try:
            return int(self.config["max_projects"])

        except KeyError:
            return 5

    def get_max_thumbnails(self):
        """Get the max thumbnails number at the data browser bottom.

        :returns: Integer
        """

        try:
            return int(self.config["max_thumbnails"])

        except KeyError:
            return 5

    def get_properties_path(self):
        """Get the path to the folder containing the processes and properties
        folders of mia (properties_path).

        The properties_path is defined and stored in the
        configuration_path.yml file, located in the ~/.populse_mia folder.

        If Mia is launched in user mode, properties path must compulsorily be
        returned from the "properties_user_path" parameter of the
        configuration_path.yml file.

        If mia is launched in developer mode, mia path must compulsorily be
        returned from the "properties_dev_path" parameter of the
        configuration_path.yml file.

        :returns: string of path to properties folder
        """
        # import verCmp only here to prevent circular import issue
        from populse_mia.utils import verCmp

        if getattr(self, "properties_path", None) is not None:
            return self.properties_path

        dot_mia_config = os.path.join(
            os.path.expanduser("~"), ".populse_mia", "configuration_path.yml"
        )

        try:
            with open(dot_mia_config, "r") as stream:
                if verCmp(yaml.__version__, "5.1", "sup"):
                    mia_home_properties_path = yaml.load(
                        stream, Loader=yaml.FullLoader
                    )

                else:
                    mia_home_properties_path = yaml.load(stream)

        except yaml.YAMLError as e:
            print(
                "\n ",
                e,
                "\n ~/.populse_mia/configuration_path.yml cannot be read, the "
                "path to the properties folder has not been found...",
            )
            raise

        else:
            # Patch for obsolete parameters.
            if "mia_path" in mia_home_properties_path:
                mia_home_properties_path["properties_user_path"] = (
                    mia_home_properties_path["mia_path"]
                )
                del mia_home_properties_path["mia_path"]

                with open(dot_mia_config, "w", encoding="utf8") as configfile:
                    yaml.dump(
                        mia_home_properties_path,
                        configfile,
                        default_flow_style=False,
                        allow_unicode=True,
                    )

            if "mia_user_path" in mia_home_properties_path:
                mia_home_properties_path["properties_user_path"] = (
                    mia_home_properties_path["mia_user_path"]
                )
                del mia_home_properties_path["mia_user_path"]

                with open(dot_mia_config, "w", encoding="utf8") as configfile:
                    yaml.dump(
                        mia_home_properties_path,
                        configfile,
                        default_flow_style=False,
                        allow_unicode=True,
                    )

            try:
                if self.dev_mode is True:
                    self.properties_path = os.path.join(
                        mia_home_properties_path["properties_dev_path"], "dev"
                    )

                elif self.dev_mode is False:
                    self.properties_path = os.path.join(
                        mia_home_properties_path["properties_user_path"], "usr"
                    )

                return self.properties_path

            except KeyError as e:
                print(
                    "\n{} key not found in ~/.populse_mia/"
                    "configuration_path.yml!\n".format(e)
                )
                raise

    def get_mri_conv_path(self):
        """Get the MRIManager.jar path.

        :returns: string of the path to the MRIManager.jar
        """

        return self.config.get("mri_conv_path", "")

    def get_mrtrix_path(self):
        """Get the  mrtrix path

        :returns: string of path to mrtrix
        """

        return self.config.get("mrtrix", "")

    def getNbAllSlicesMax(self):
        """Get number the maximum number of slices to display in the
        miniviewer.

        :returns: Integer
        """

        return int(self.config.get("nb_slices_max", "10"))

    def get_opened_projects(self):
        """Get opened projects.

        :returns: list of opened projects
        """

        return self.config.get("opened_projects", [])

    def getPathToProjectsFolder(self):
        """Get the project's path.

        :returns: string of the path
        """

        return self.config.get("projects_save_path", "")

    def get_projects_save_path(self):
        """Get the path where projects are saved.

        :returns: string of path
        """

        try:
            return self.config["projects_save_path"]

        except KeyError:
            # if not os.path.isdir(
            #         os.path.join(self.get_properties_path(), 'projects')):
            #     os.mkdir(os.path.join(self.get_properties_path(),
            #                           'projects'))
            #
            # return os.path.join(self.get_properties_path(), 'projects')
            return ""

    def get_referential(self):
        """Checks in anatomist_2 data viewer which referential has been chosen

        :returns: 0 for World Coordinates, 1 for Image ref
        """

        return self.config.get("ref", "0")

    def get_resources_path(self):
        """Get the resources path.

        :returns: string of the path to the resources folder
        """

        return self.config.get("resources_path", "")

    def getShowAllSlices(self):
        """Get whether the show_all_slices parameters was enabled
        or not in the miniviewer.

        :returns: boolean
        """

        # Used in MiniViewer
        return self.config.get("show_all_slices", False)

    def getSourceImageDir(self):
        """Get the source directory for project images.

        :returns: string of the path
        """

        return self.config.get("source_image_dir", "")

    def get_spm_path(self):
        """Get the path of SPM12.

        :returns: string of path
        """

        return self.config.get("spm", "")

    def get_spm_standalone_path(self):
        """Get the path to the SPM12 (standalone version).

        :returns: String of path
        """

        return self.config.get("spm_standalone", "")

    def getTextColor(self):
        """Get the text color.

        :returns: string
        """

        return self.config.get("text_color", "")

    def getThumbnailTag(self):
        """Get the tag of the thumbnail displayed in the miniviewer.

        :returns: string
        """

        return self.config.get("thumbnail_tag", "SequenceName")

    def get_use_afni(self):
        """Get the value of "use afni" checkbox in the preferences.

        :returns: boolean
        """

        return self.config.get("use_afni", False)

    def get_use_ants(self):
        """Get the value of "use ants" checkbox in the preferences.

        :returns: boolean
        """

        return self.config.get("use_ants", False)

    def get_use_clinical(self):
        """Get the clinical mode in the preferences.

        :returns: boolean
        """

        return self.config.get("clinical_mode", False)

    def get_use_fsl(self):
        """Get the value of "use fsl" checkbox in the preferences.

        :returns: boolean
        """

        return self.config.get("use_fsl", False)

    def get_use_freesurfer(self):
        """Get the value of "use freesurfer" checkbox in the preferences.

        :returns: boolean
        """

        return self.config.get("use_freesurfer", False)

    def get_use_matlab(self):
        """Get the value of "use matlab" checkbox in the preferences.

        :returns: boolean
        """

        return self.config.get("use_matlab", False)

    def get_use_matlab_standalone(self):
        """Get the value of "use matlab standalone" checkbox in the
        preferences.

        :returns: boolean
        """

        return self.config.get("use_matlab_standalone", False)

    def get_user_level(self):
        """Get the user level in the Capsul config

        :returns: integer
        """

        return (
            self.config.get("capsul_config", {})
            .get("engine", {})
            .get("global", {})
            .get("capsul.engine.module.axon", {})
            .get("user_level", 0)
        )

    def get_user_mode(self):
        """Get if user mode is disabled or enabled in the preferences.

        :returns: boolean
        """

        return self.config.get("user_mode", True)

    def get_use_mrtrix(self):
        """Get the value of "use mrtrix" checkbox in the preferences.

        :returns: boolean
        """

        return self.config.get("use_mrtrix", False)

    def get_use_spm(self):
        """Get the value of "use spm" checkbox in the preferences.

        :returns: boolean
        """

        return self.config.get("use_spm", False)

    def get_use_spm_standalone(self):
        """Get the value of "use spm standalone" checkbox in the preferences.

        :returns: boolean
        """

        return self.config.get("use_spm_standalone", False)

    def getViewerConfig(self):
        """Get the viewer config neuro or radio, neuro by default

        :returns: String
        """

        return self.config.get("config_NeuRad", "neuro")

    def getViewerFramerate(self):
        """Get the Viewer framerate

        :returns: integer
        """

        return self.config.get("im_sec", "5")

    def isAutoSave(self):
        """Get if the auto-save mode is enabled or not.

        :returns: boolean
        """

        return self.config.get("auto_save", False)

    def isControlV1(self):
        """Get if the display of the controller is of V1 type.

        :returns: boolean
        """

        return self.config.get("control_V1", False)

    def isRadioView(self):
        """Get if the display in miniviewer is in radiological orientation.

        - True for radiological
        - False for neurological

        :returns: boolean
        """

        return self.config.get("radio_view", True)

    def loadConfig(self):
        """Read the config in the config.yml file.

        :returns: Returns a dictionary of the contents of config.yml
        """

        # import verCmp only here to prevent circular import issue
        from populse_mia.utils import verCmp

        f = Fernet(CONFIG)
        config_file = os.path.join(
            self.get_properties_path(), "properties", "config.yml"
        )

        if not os.path.exists(config_file):
            raise yaml.YAMLError(
                "\nThe '{}' file doesn't exist or is "
                "corrupted...\n".format(config_file)
            )

        with open(config_file, "rb") as stream:
            try:
                stream = b"".join(stream.readlines())
                decrypted = f.decrypt(stream)

                if verCmp(yaml.__version__, "5.1", "sup"):
                    return yaml.load(decrypted, Loader=yaml.FullLoader)

                else:
                    return yaml.load(decrypted)

            except yaml.YAMLError as exc:
                print("error loading YAML file: %s" % config_file)
                print(exc)

        # in case of problem, return an empty config
        return {}

    def saveConfig(self):
        """Save the current parameters in the config.yml file."""

        f = Fernet(CONFIG)
        config_file = os.path.join(
            self.get_properties_path(), "properties", "config.yml"
        )

        if not os.path.exists(os.path.dirname(config_file)):
            os.makedirs(os.path.dirname(config_file))

        with open(config_file, "wb") as configfile:
            stream = yaml.dump(
                self.config, default_flow_style=False, allow_unicode=True
            )
            configfile.write(f.encrypt(stream.encode()))

        self.update_capsul_config()

    def set_admin_hash(self, admin_hash):
        """Set the password hash.

        :param admin_hash: string of hash
        """

        self.config["admin_hash"] = admin_hash
        # Then save the modification
        self.saveConfig()

    def set_afni_path(self, path):
        """Set the AFNI path

        :param path: string of AFNI path
        """

        self.config["afni"] = path
        # Then save the modification
        self.saveConfig()

    def set_ants_path(self, path):
        """Set the ANTS path

        :param path: string of ANTS path
        """

        self.config["ants"] = path
        # Then save the modification
        self.saveConfig()

    def set_mrtrix_path(self, path):
        """Set the mrtrix path

        :param path: string of mrtrix path
        """

        self.config["mrtrix"] = path
        # Then save the modification
        self.saveConfig()

    # def set_freesurfer_path(self, path):
    #     """Set the freesurfer path

    #     :param path: string of freesurfer path
    #     """

    #     self.config["freesurfer"] = path
    #     # Then save the modification
    #     self.saveConfig()

    def setAutoSave(self, save):
        """Set auto-save mode.

        :param save: boolean
        """

        self.config["auto_save"] = save
        # Then save the modification
        self.saveConfig()

    def setBackgroundColor(self, color):
        """Set background color and save configuration.

        :param color: Color string ('Black', 'Blue', 'Green', 'Grey',
                                    'Orange', 'Red', 'Yellow', 'White')
        """

        self.config["background_color"] = color
        # Then save the modification
        self.saveConfig()

    def set_capsul_config(self, capsul_config_dict):
        """Set CAPSUL configuration dict into MIA config.

        This method is used just (and only) after editing capsul config
        (in File > Mia preferences, Pipeline tab, Edit CAPSUL config button),
        in order to synchronise the new Capsul config with the Mia preferences.

        :param capsul_config_dict: a dict; {'engine': {...},
                                            'engine_modules': [...]}
        """

        self.config["capsul_config"] = capsul_config_dict

        # update MIA values
        engine_config = capsul_config_dict.get("engine")
        new_engine = capsul_engine()

        for environment, config in engine_config.items():
            if environment == "capsul_engine":
                continue
            new_engine.import_configs(environment, config)

        engine_config = new_engine.settings.export_config_dict("global")

        # afni
        afni = engine_config.get("global", {}).get("capsul.engine.module.afni")

        if afni:
            afni = next(iter(afni.values()))
            afni_path = afni.get("directory")
            use_afni = bool(afni_path)

            if afni_path:
                self.set_afni_path(afni_path)

            self.set_use_afni(use_afni)

        # ants
        ants = engine_config.get("global", {}).get("capsul.engine.module.ants")

        if ants:
            ants = next(iter(ants.values()))
            ants_path = ants.get("directory")
            use_ants = bool(ants_path)

            if ants_path:
                self.set_ants_path(ants_path)

            self.set_use_ants(use_ants)

        # freesurfer
        freesurfer = engine_config.get("global", {}).get(
            "capsul.engine.module.freesurfer"
        )
        use_freesurfer = False
        if freesurfer:
            freesurfer = next(iter(freesurfer.values()))
            freesurfer_setup_path = freesurfer.get("setup")

            if freesurfer_setup_path:
                use_freesurfer = True
                self.set_freesurfer_setup(freesurfer_setup_path)

            self.set_use_freesurfer(use_freesurfer)

        # fsl
        fsl = engine_config.get("global", {}).get("capsul.engine.module.fsl")
        use_fsl = False

        if fsl:
            fsl = next(iter(fsl.values()))
            fsl_conf_path = fsl.get("config")
            fsl_dir_path = fsl.get("directory")

            if fsl_conf_path:
                use_fsl = True
                self.set_fsl_config(fsl_conf_path)
                self.set_use_fsl(True)

            # If only the directory parameter has been set, let's try using
            # the config parameter = directory/etc/fslconf/fsl.sh:
            elif fsl_dir_path:
                fsl_conf = os.path.join(
                    fsl_dir_path, "etc", "fslconf", "fsl.sh"
                )

                if os.path.isfile(fsl_conf):
                    use_fsl = True
                    self.set_fsl_config(fsl_conf)
                    self.set_use_fsl(True)

                else:
                    print(
                        "\nThe automatic determination of the configuration "
                        "file from the directory known for fsl ({}) did not "
                        "work. FSL is not correctly defined in the "
                        "preferences (see File > Mia Preferences) "
                        "...".format(fsl_dir_path)
                    )

        if use_fsl is False:
            self.set_use_fsl(False)

        # matlab
        matlab = engine_config.get("global", {}).get(
            "capsul.engine.module.matlab"
        )
        use_matlab = False
        use_mcr = False

        if matlab:
            matlab = next(iter(matlab.values()))
            matlab_path = matlab.get("executable")

            if bool(matlab_path) and os.path.isfile(matlab_path):
                use_matlab = True

            mcr_dir = matlab.get("mcr_directory")

            if bool(mcr_dir) and os.path.isdir(mcr_dir):
                use_mcr = True

        # mrtrix
        mrtrix = engine_config.get("global", {}).get(
            "capsul.engine.module.mrtrix"
        )

        if mrtrix:
            mrtrix = next(iter(mrtrix.values()))
            mrtrix_path = mrtrix.get("directory")
            use_mrtrix = bool(mrtrix_path)

            if mrtrix_path:
                self.set_mrtrix_path(mrtrix_path)

            self.set_use_mrtrix(use_mrtrix)

        # spm
        spm = engine_config.get("global", {}).get("capsul.engine.module.spm")
        use_spm_standalone = False
        use_spm = False

        if spm:
            # TODO: we only take the first element of the dictionary (the one
            #       that is normally edited in the Capsul config GUI). There is
            #       actually a problem because this means that there may be
            #       hidden config(s) ... This can produce bugs and at least
            #       unpredictable results for the user ...

            spm = next(iter(spm.values()))
            spm_dir = spm.get("directory", False)
            use_spm_standalone = spm.get("standalone", False)

            if use_spm_standalone and os.path.isdir(spm_dir) and use_mcr:
                pass

            else:
                use_spm_standalone = False

            if (
                use_spm_standalone is False
                and os.path.isdir(spm_dir)
                and use_matlab
            ):
                use_spm = True

            else:
                use_spm = False

        if use_spm:
            self.set_spm_path(spm_dir)
            self.set_use_spm(True)
            self.set_use_spm_standalone(False)
            self.set_matlab_path(matlab_path)
            self.set_use_matlab(True)
            self.set_use_matlab_standalone(False)

        elif use_spm_standalone:
            self.set_spm_standalone_path(spm_dir)
            self.set_use_spm_standalone(True)
            self.set_use_spm(False)
            self.set_matlab_standalone_path(mcr_dir)
            self.set_use_matlab_standalone(True)
            self.set_use_matlab(False)

        # TODO: Because there are two parameters for matlab (executable and
        #  mcr_directory) in Capsul config, if the user defines both, we don't
        #  know which on to choose! Here we choose to favour matlab in front
        #  of MCR if both are chosen, is it desirable?
        elif use_matlab:
            self.set_matlab_path(matlab_path)
            self.set_use_matlab(True)
            self.set_use_matlab_standalone(False)
            self.set_use_spm(False)
            self.set_use_spm_standalone(False)

        elif use_mcr:
            self.set_matlab_standalone_path(mcr_dir)
            self.set_use_matlab_standalone(True)
            self.set_use_matlab(False)
            self.set_use_spm(False)
            self.set_use_spm_standalone(False)

        else:
            self.set_use_matlab(False)
            self.set_use_matlab_standalone(False)
            self.set_use_spm(False)
            self.set_use_spm_standalone(False)

        if (
            use_matlab
            and use_mcr
            and use_spm is False
            and use_spm_standalone is False
        ):
            print(
                "\n The Matlab executable and the mcr_directory parameters "
                "have been set concomitantly in the Capsul configuration. "
                "This leads to an indeterminacy. By default, Matlab is "
                "retained at the expense of MCR."
            )

        self.update_capsul_config()  # store into capsul engine

    def setChainCursors(self, chain_cursors):
        """Set the value of the checkbox 'chain cursor' in the mini viewer.

        :param chain_cursors: Boolean
        """

        self.config["chain_cursors"] = chain_cursors
        # Then save the modification
        self.saveConfig()

    def set_clinical_mode(self, clinical_mode):
        """Enable of disable clinical mode.

        :param clinical_mode: boolean
        """

        self.config["clinical_mode"] = clinical_mode
        # Then save the modification
        self.saveConfig()

    def setControlV1(self, controlV1):
        """Set controller display mode (True if V1).

        :param controlV1: boolean
        """

        self.config["control_V1"] = controlV1
        # Then save the modification
        self.saveConfig()

    def set_fsl_config(self, path):
        """Set  the FSL config file

        :param path: string of path to fsl/etc/fslconf/fsl.sh
        """

        self.config["fsl_config"] = path
        # Then save the modification
        self.saveConfig()

    def set_freesurfer_setup(self, path):
        """Set  the freesurfer config file

        :param path: string of path to freesurfer/FreeSurferEnv.sh
        """

        self.config["freesurfer_setup"] = path
        # Then save the modification
        self.saveConfig()

    def set_mainwindow_maximized(self, enabled):
        """Set the maximized (full-screen) flag

        :param enabled: boolean
        """

        self.config["mainwindow_maximized"] = enabled
        self.saveConfig()

    def set_mainwindow_size(self, size):
        """Set main window size

        :param size: list
        """

        self.config["mainwindow_size"] = list(size)
        self.saveConfig()

    def set_matlab_path(self, path):
        """Set the path of Matlab's executable.

        :param path: string of path
        """

        self.config["matlab"] = path
        # Then save the modification
        self.saveConfig()

    def set_matlab_standalone_path(self, path):
        """Set the path of Matlab Compiler Runtime.

        :param path: string of path
        """

        self.config["matlab_standalone"] = path
        # Then save the modification
        self.saveConfig()

    def set_max_projects(self, nb_max_projects):
        """Set the maximum number of projects displayed in
        the "Saved projects" menu.

        :param nb_max_projects: Integer
        """

        self.config["max_projects"] = nb_max_projects
        # Then save the modification
        self.saveConfig()

    def set_max_thumbnails(self, nb_max_thumbnails):
        """Set max thumbnails number at the data browser bottom.

        :param nb_max_thumbnails: Integer
        """

        self.config["max_thumbnails"] = nb_max_thumbnails
        # Then save the modification
        self.saveConfig()

    def set_mri_conv_path(self, path):
        """Set the MRIManager.jar path.

        :param path: string of the path
        """

        self.config["mri_conv_path"] = path
        # Then save the modification
        self.saveConfig()

    def setNbAllSlicesMax(self, nb_slices_max):
        """Set the number of slices to display in the mini viewer.

        :param nb_slices_max: maximum number of slices to display (Int)
        """

        self.config["nb_slices_max"] = nb_slices_max
        # Then save the modification
        self.saveConfig()

    def set_opened_projects(self, new_projects):
        """Set the list of opened projects and saves the modification.

        :param new_projects: list of path
        """

        self.config["opened_projects"] = new_projects
        # Then save the modification
        self.saveConfig()

    def set_projects_save_path(self, path):
        """Set the folder where the projects are saved.

        :param path: string of path
        """

        self.config["projects_save_path"] = path
        # Then save the modification
        self.saveConfig()

    def set_radioView(self, radio_view):
        """Set the radiological/neurological orientation in mini viewer.

        - True for radiological
        - False for neurological

        :param radio_view: boolean
        """

        self.config["radio_view"] = radio_view
        # Then save the modification
        self.saveConfig()

    def set_referential(self, ref):
        """Set the referential to image ref or world coordinates in anatomist_2
        data viewer.

        :param ref: str; 0 for World Coordinates, 1 for Image ref
        """

        self.config["ref"] = ref
        # Then save the modification
        self.saveConfig()

    def set_resources_path(self, path):
        """Set the resources path.

        :param path: string of the path
        """

        self.config["resources_path"] = path
        # Then save the modification
        self.saveConfig()

    def setShowAllSlices(self, show_all_slices):
        """Set the show_all_slides setting in miniviewer.

        :param show_all_slices: Boolean
        """

        self.config["show_all_slices"] = show_all_slices
        # Then save the modification
        self.saveConfig()

    def setSourceImageDir(self, source_image_dir):
        """Set the source directory for project images.

        :param source_image_dir: String of path
        """

        self.config["source_image_dir"] = source_image_dir
        # Then save the modification
        self.saveConfig()

    def set_spm_path(self, path):
        """Set the path of SPM (license version).

        :param path: string of path
        """

        self.config["spm"] = path
        # Then save the modification
        self.saveConfig()

    def set_spm_standalone_path(self, path):
        """Set the path of SPM (standalone version).

        :param path: string of path
        """

        self.config["spm_standalone"] = path
        # Then save the modification
        self.saveConfig()

    def setTextColor(self, color):
        """Set text color and save configuration.

        :param color: Color string ('Black', 'Blue', 'Green', 'Grey',
                                    'Orange', 'Red', 'Yellow', 'White')
        """

        self.config["text_color"] = color
        # Then save the modification
        self.saveConfig()

    def setThumbnailTag(self, thumbnail_tag):
        """Set the tag that is displayed in the mini viewer.

        :param thumbnail_tag: string
        """

        self.config["thumbnail_tag"] = thumbnail_tag
        # Then save the modification
        self.saveConfig()

    def set_use_afni(self, use_afni):
        """Set the value of "use_afni" checkbox in the preferences.

        :param use_afni: boolean
        """

        self.config["use_afni"] = use_afni
        # Then save the modification
        self.saveConfig()

    def set_use_ants(self, use_ants):
        """Set the value of "use_ants" checkbox in the preferences.

        :param use_ants: boolean
        """

        self.config["use_ants"] = use_ants
        # Then save the modification
        self.saveConfig()

    def set_use_freesurfer(self, use_freesurfer):
        """Set the value of "use_freesurfer" checkbox in the preferences.

        :param use_freesurfer: boolean
        """

        self.config["use_freesurfer"] = use_freesurfer
        # Then save the modification
        self.saveConfig()

    def set_use_fsl(self, use_fsl):
        """Set the value of "use_fsl" checkbox in the preferences.

        :param use_fsl: boolean
        """

        self.config["use_fsl"] = use_fsl
        # Then save the modification
        self.saveConfig()

    def set_use_matlab(self, use_matlab):
        """Set the value of "use matlab" checkbox in the preferences.

        :param use_matlab: boolean
        """

        self.config["use_matlab"] = use_matlab
        # Then save the modification
        self.saveConfig()

    def set_use_matlab_standalone(self, use_matlab_standalone):
        """Set the value of "use_matlab_standalone" checkbox in the
        preferences.

        :param use_matlab: boolean
        """

        self.config["use_matlab_standalone"] = use_matlab_standalone
        # Then save the modification
        self.saveConfig()

    def set_use_mrtrix(self, use_mrtrix):
        """Set the value of "use_mrtrix" checkbox in the preferences.

        :param use_mrtrix: boolean
        """

        self.config["use_mrtrix"] = use_mrtrix
        # Then save the modification
        self.saveConfig()

    def set_user_mode(self, user_mode):
        """Enable of disable user mode.

        :param user_mode: boolean
        """

        self.config["user_mode"] = user_mode
        # Then save the modification
        self.saveConfig()

    def set_use_spm(self, use_spm):
        """Set the value of "use spm" checkbox in the preferences.

        :param use_spm: boolean
        """

        self.config["use_spm"] = use_spm
        # Then save the modification
        self.saveConfig()

    def set_use_spm_standalone(self, use_spm_standalone):
        """Set the value of "use spm standalone" checkbox in the preferences.

        :param use_spm_standalone: boolean
        """

        self.config["use_spm_standalone"] = use_spm_standalone
        # Then save the modification
        self.saveConfig()

    def setViewerConfig(self, config_NeuRad):
        """sets user's configuration neuro or radio for data_viewer

        - neuro: neurological
        - radio: radiological

        :param config_NeuRad: string
        """

        self.config["config_NeuRad"] = config_NeuRad
        # Then save the modification
        self.saveConfig()

    def setViewerFramerate(self, im_sec):
        """sets user's framerate for data_viewer

        :param im_sec: int
        """

        self.config["im_sec"] = im_sec
        # Then save the modification
        self.saveConfig()

    def update_capsul_config(self):
        """
        Update a global CapsulEngine object used for all operations in MIA
        application. The engine is created once when needed, and updated
        each time the config is saved.

        :returns: capsul.engine.CapsulEngine object
        """

        if self.capsul_engine is None:
            # don't do anything until the config is really created: this
            # avoids unneeded updates before it is actually used.
            return

        capsul_config = self.get_capsul_config(sync_from_engine=False)
        engine = Config.capsul_engine

        for module in capsul_config.get("engine_modules", []) + [
            "fom",
            "nipype",
            "python",
            "fsl",
            "freesurfer",
            "axon",
            "afni",
            "ants",
            "mrtrix",
            "somaworkflow",
        ]:
            engine.load_module(module)

        engine_config = capsul_config.get("engine")

        if engine_config:
            for environment, config in engine_config.items():
                c = dict(config)

                if (
                    "capsul_engine" not in c
                    or "uses" not in c["capsul_engine"]
                ):
                    c["capsul_engine"] = {
                        "uses": {
                            engine.settings.module_name(m): "ALL"
                            for m in config.keys()
                        }
                    }

                try:
                    engine.import_configs(environment, c, cont_on_error=True)

                except Exception as exc:
                    print(
                        "\nAn issue is detected in the Mia's "
                        "configuration:\n{}\nPlease check the settings "
                        "in File > Mia Preferences > Pipeline "
                        "...".format(exc)
                    )

        return engine



# -*- coding: utf-8 -*-
"""Module that contains the class to handle the projects saved in the software.

Contains:
    Class:
    - SavedProjects

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import os

import yaml
from packaging import version

# Populse_MIA imports
from populse_mia.software_properties import Config


class SavedProjects:
    """Class that handles all the projects that have been saved in the
    software.

    .. Methods:
        - addSavedProject: adds a new saved project
        - loadSavedProjects: loads the dictionary from the saved_projects.yml
        file
        - removeSavedProject: removes a saved project from the config file
        - saveSavedProjects: saves the dictionary to the saved_projects.yml
        file


    """

    def __init__(self):
        """Initialise the savedProjects attribute from the saved_projects.yml
           file.

        The pathsList attribute is initialised as the value corresponding to
        the "paths" key in the savedProjects dictionary.

        """
        self.savedProjects = self.loadSavedProjects()

        if (isinstance(self.savedProjects, dict)) and (
            "paths" in self.savedProjects
        ):
            self.pathsList = self.savedProjects["paths"]

            if self.pathsList is None:
                self.pathsList = []

        else:
            self.savedProjects = {"paths": []}
            self.pathsList = []

    def addSavedProject(self, newPath):
        """Add a new project to save in the savedProjects and pathsList
        attributes.

        Finally, save the savedProjects attribute in the saved_projects.yml
        file.

        :param newPath: new project's path to add

        :returns: the new path's list (pathsList attribute)

        """
        if self.pathsList:
            if newPath not in self.pathsList:
                self.pathsList.insert(0, newPath)

            elif newPath != self.pathsList[0]:
                self.pathsList.remove(newPath)
                self.pathsList.insert(0, newPath)

        else:
            self.pathsList.insert(0, newPath)

        self.savedProjects["paths"] = self.pathsList
        self.saveSavedProjects()
        return self.pathsList

    def loadSavedProjects(self):
        """Load the savedProjects dictionary from the saved_projects.yml file.

        If the saved_projects.yml file is not existing, it is created with the
        "{paths: []}" value and the returned dictionary is {paths: []}.

        :returns: the dictionary

        """
        config = Config()

        try:
            with open(
                os.path.join(
                    config.get_properties_path(),
                    "properties",
                    "saved_projects.yml",
                ),
                "r",
            ) as stream:
                try:
                    if version.parse(yaml.__version__) > version.parse("5.1"):
                        return yaml.load(stream, Loader=yaml.FullLoader)
                    else:
                        return yaml.load(stream)

                except yaml.YAMLError as exc:
                    print(exc)

        except FileNotFoundError:
            with open(
                os.path.join(
                    config.get_properties_path(),
                    "properties",
                    "saved_projects.yml",
                ),
                "w",
            ) as stream:
                yaml.dump({"paths": []}, stream, default_flow_style=False)
                return {"paths": []}

    def removeSavedProject(self, path):
        """Removes a saved project from the saved_projects.yml file.

        :param path: path of the saved project to remove

        """

        if path in self.savedProjects["paths"]:
            self.savedProjects["paths"].remove(path)

        self.saveSavedProjects()

    def saveSavedProjects(self):
        """Saves the savedProjects dictionary to the saved_projects.yml
        file."""

        config = Config()

        with open(
            os.path.join(
                config.get_properties_path(),
                "properties",
                "saved_projects.yml",
            ),
            "w",
            encoding="utf8",
        ) as configfile:
            yaml.dump(
                self.savedProjects,
                configfile,
                default_flow_style=False,
                allow_unicode=True,
            )
